import logging
from collections import namedtuple, defaultdict
from pathlib import Path
from random import random
import numpy as np
import toml
from operator import itemgetter
import json
import jsonschema
from enum import IntEnum

from ru.channels import MINION_CHANNELS, FLONGLE_CHANNELS


class Severity(IntEnum):
    INFO = 1
    WARN = 2
    ERROR = 3


def send_message(rpc_connection, message, severity):
    """Send a message to MinKNOW

    Parameters
    ----------
    rpc_connection
        An instance of the rpc.Connection
    message : str
        The message to send
    severity : int
        The severity to use for the message: 1=info, 2=warning, 3=error

    Returns
    -------
    None
    """
    rpc_connection.log.send_user_message(severity=severity, user_message=message)


def dynamic_import(name):
    """Dynamically import modules and classes, used to get the ReadCache

    https://stackoverflow.com/a/547867/3279716
    https://docs.python.org/2.4/lib/built-in-funcs.html

    Parameters
    ----------
    name : str
        The module/class path. E.g: "read_until.read_cache.{}".format("ReadCache")

    Returns
    -------
    module
    """
    components = name.split('.')
    mod = __import__(components[0])
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod


def named_tuple_generator(dictionary, name='Conditions',):
    """Generate named tuple from dictionary

    Parameters
    ----------
    dictionary : dict
        dict to turn into a named tuple
    name : str
        The name to give the named tuple

    Returns
    -------
    namedtuple
    """
    return namedtuple(name, dictionary.keys())(**dictionary)


def nice_join(seq, sep=", ", conjunction="or"):
    """Join lists nicely"""
    seq = [str(x) for x in seq]

    if len(seq) <= 1 or conjunction is None:
        return sep.join(seq)
    else:
        return "{} {} {}".format(sep.join(seq[:-1]), conjunction, seq[-1])


def get_log_level(s):
    """Get log level from logging"""
    return getattr(logging, s.upper())


def read_lines_to_list(f):
    """Read file to list and return the list"""
    with open(f) as fh:
        lines = [line.strip() for line in fh]
    return lines


def print_args(args, logger=None, exclude=None):
    """Print and format all arguments from the command line"""
    if exclude is None:
        exclude = []
    dirs = dir(args)
    m = max([len(a) for a in dirs if a[0] != '_'])
    for attr in dirs:
        if attr[0] != '_' and attr not in exclude and attr.lower() == attr:
            if logger is not None:
                logger.info("{a}={b}".format(a=attr, m=m, b=getattr(args, attr)))
            else:
                print('{a:<{m}}\t{b}'.format(a=attr, m=m, b=getattr(args, attr)))


def get_coords(channel, flowcell_size):
    """Return a channel's coordinates given a flowcell size

    Parameters
    ----------
    channel : int
        The channel to retrieve the coordinates for
    flowcell_size : int
        The flowcell size, this is used to determine the flowcell layout

    Returns
    -------
    tuple
        Tuple of int: (column, row)

    Raises
    ------
    ValueError
        Raised if channel outside of bounds (0, flowcell_size)
        Raised if flowcell_size not one of [128, 512, 3000]
    """
    if channel <= 0 or channel > flowcell_size:
        raise ValueError("channel cannot be below 0 or above flowcell_size")

    if flowcell_size == 3000:
        # find which block of 12 we are in:
        block = (channel - 1) // 250
        remainder = (channel - 1) % 250
        row = remainder // 10
        column = remainder % 10 + block * 10
        return column, row
    elif flowcell_size == 126:
        return FLONGLE_CHANNELS[channel]
    elif flowcell_size == 512:
        return MINION_CHANNELS[channel]
    else:
        raise ValueError("flowcell_size is not recognised")


def get_flowcell_array(flowcell_size):
    """Return a numpy.ndarray in the shape of a flowcell

    Parameters
    ----------
    flowcell_size : int
        The total number of channels on the flowcell; 126 for Flongle, 512
        for MinION, and 3000 for PromethION

    Returns
    -------
    np.ndarray
        An N-dimensional array representation of the flowcell

    Examples
    --------
    >>> get_flowcell_array(126).shape
    (10, 13)
    >>> get_flowcell_array(512).shape
    (16, 32)
    >>> get_flowcell_array(3000).shape
    (25, 120)
    >>> get_flowcell_array(128)
    Traceback (most recent call last):
        ...
    ValueError: flowcell_size is not recognised
    >>> get_flowcell_array(126)[9][-1]
    0
    >>> get_flowcell_array(512)[15][-1]
    1

    """
    # Make a list of tuples of (column, row, channel)
    coords = [(*get_coords(x, flowcell_size), x) for x in range(1, flowcell_size + 1)]

    # Initialise a nd array using the max row and column from coords
    b = np.zeros(
        (
            max(coords, key=itemgetter(1))[1] + 1,
            max(coords, key=itemgetter(0))[0] + 1
        ),
        dtype=int
    )

    # Mimic flowcell layout in an array
    for col, row, chan in coords:
        b[row][col] += chan

    # return the reversed array, to get the right orientation
    return b[::-1]


def generate_flowcell(flowcell_size, split=1, axis=1, odd_even=False):
    """Return an list of lists with channels to use in conditions

    Representations generated by this method are evenly split based on the physical
    layout of the flowcell. Each sub-list is the same size. Axis determines whether
    the flowcell divisions will go left-right (0) or top-bottom (1); as flongle has
    a shape of (10, 13) the top-bottom axis cannot be split evenly.

    Parameters
    ----------
    flowcell_size : int
        The total number of channels on the flowcell; 126 for Flongle, 512 for MinION,
        and 3000 for PromethION
    split : int
        The number of sections to split the flowcell into, must be a positive factor
        of the flowcell dimension
    axis : int, optional
        The axis along which to split,
        see: https://docs.scipy.org/doc/numpy/glossary.html?highlight=axis
    odd_even : bool
        Return a list of two lists split into odd-even channels,
        ignores `split` and `axis`

    Returns
    -------
    list
        A list of lists with channels divided equally

    Raises
    ------
    ValueError
        Raised when split is not a positive integer
        Raised when the value for split is not a factor on the axis provided

    Examples
    --------
    >>> len(generate_flowcell(512))
    1
    >>> len(generate_flowcell(512)[0])
    512
    >>> len(generate_flowcell(512, split=4))
    4
    >>> for x in generate_flowcell(512, split=4):
    ...     print(len(x))
    128
    128
    128
    128
    >>> generate_flowcell(512, split=5)
    Traceback (most recent call last):
        ...
    ValueError: The flowcell cannot be split evenly
    """
    if odd_even:
        return [
            list(range(1, flowcell_size + 1, 2)),
            list(range(2, flowcell_size + 1, 2)),
        ]

    arr = get_flowcell_array(flowcell_size)

    if split <= 0:
        raise ValueError("split must be a positive integer")

    try:
        arr = np.array(np.split(arr, split, axis=axis))
    except ValueError:
        # The number of targets cannot be split evenly over the flowcell.
        #   For MinION flowcells the number of targets must be a factor of 16 or
        #   32 for axis 0 or 1 respectively; for PromethION flowcells the number
        #   of targets must be a factor of 25 or 120 for axis 0 or 1 respectively.
        raise ValueError("The flowcell cannot be split evenly")

    arr.shape = (arr.shape[0], arr.shape[1] * arr.shape[2])
    return [x for x in arr.tolist()]


def get_targets(targets):
    """

    Parameters
    ----------
    targets : str or List[str]

    Returns
    -------
    defaultdict of list
    {
        'strand':
            {
                'contig': [(int, int), ...]
            }
    }
    """
    t = defaultdict(lambda: defaultdict(list))
    if isinstance(targets, str):
        # Load from list
        if Path(targets).is_file():
            targets = read_lines_to_list(targets)
        # If targets is not a file, then raise error

    for item in targets:
        ctg, *coords = item.split(",")
        if coords:
            strand = coords.pop()
            t[strand][ctg].append(tuple(int(x) for x in coords))
        else:
            for strand in ["+", "-"]:
                t[strand][ctg].append((0, float("inf")))

    return t


def load_config_toml(filepath, validate=True):
    """Load a TOML file and check file paths

    Parameters
    ----------
    filepath : str
        Path to the TOML config file
    validate : bool
        If True, test TOML file against the JSON schema (./static/readfish_toml.schema.json)

    Returns
    -------
    dict
        Returns dict of TOML config
    """
    # Check that TOML config file exists
    p = Path(filepath)
    if not p.is_file():
        raise FileNotFoundError("TOML config file not found at '{}'".format(filepath))

    # Load TOML to dict
    toml_dict = toml.load(p)

    # Check reference path
    reference_text = toml_dict.get("conditions", {}).get("reference", "")
    reference_path = Path(reference_text)
    if not reference_path.is_file() and reference_text:
        raise FileNotFoundError("Reference file not found at '{}'".format(reference_path))

    # Get keys for all condition tables, allows safe updates
    conditions = [k for k, cond in toml_dict.get("conditions", {}).items() if isinstance(cond, dict)]

    # Load targets from a file
    for k in conditions:
        targets = toml_dict["conditions"][k].get("targets", [])
        if isinstance(targets, str):
            if not Path(targets).is_file():
                raise FileNotFoundError("Targets file not found at '{}'".format(targets))

            toml_dict["conditions"][k]["targets"] = read_lines_to_list(targets)

    # Validate our TOML file
    if validate:
        # Load json schema
        _f = Path(__file__).parent / "static/readfish_toml.schema.json"
        with _f.resolve().open() as fh:
            schema = json.load(fh)
        try:
            jsonschema.validate(toml_dict, schema)
        except jsonschema.exceptions.ValidationError as err:
            print("😾 this TOML file has failed validation. See below for details:")
            raise

    return toml_dict


def describe_experiment(conditions, mapper):
    """

    Parameters
    ----------
    conditions : List[NamedTuple, ...]
        List of named tuples, should be conditions from get_run_info
    mapper : mappy.mapper
        Instance of mappy.mapper initialised with the reference passed from
        get_run_info

    Yields
    ------
    str
        Message string
    severity : Severity
        One of Severity.INFO, Severity.WARN or Severity.ERROR

    """
    # TODO: conditional 's' here
    yield "This experiment has {} region{} on the flowcell".format(
        len(conditions), {1: ""}.get(len(conditions), "s")
    ), Severity.INFO

    if mapper.initialised:
        yield "Using reference: {}".format(mapper.index), Severity.INFO
        seq_names = set(mapper.mapper.seq_names)

        for region in conditions:
            conds = {
                "unblock": [],
                "stop_receiving": [],
                "proceed": [],
            }
            for m in ("single_on", "single_off", "multi_on", "multi_off", "no_map", "no_seq"):
                conds[getattr(region, m)].append(m)
            conds = {k: nice_join(v) for k, v in conds.items()}
            s = (
                "Region '{}' (control={}) has {} target{} of which {} are in the reference. "
                "Reads will be unblocked when classed as {unblock}; sequenced when classed as "
                "{stop_receiving}; and polled for more data when classed as {proceed}.".format(
                    region.name,
                    region.control,
                    len(region.targets),
                    {1: ""}.get(len(region.targets), "s"),
                    len(region.targets & seq_names),
                    **conds,
                )
            )
            yield s, Severity.INFO
    else:
        yield "No reference file provided", Severity.WARN
        for region in conditions:
            conds = {
                "unblock": [],
                "stop_receiving": [],
                "proceed": [],
            }
            for m in ("single_on", "single_off", "multi_on", "multi_off", "no_map", "no_seq"):
                conds[getattr(region, m)].append(m)
            conds = {k: nice_join(v) for k, v in conds.items()}
            s = (
                "Region '{}' (control={}) has {} target{}. "
                "Reads will be unblocked when classed as {unblock}; sequenced when classed as "
                "{stop_receiving}; and polled for more data when classed as {proceed}.".format(
                    region.name,
                    region.control,
                    len(region.targets),
                    {1: ""}.get(len(region.targets), "s"),
                    **conds,
                )
            )
            yield s, Severity.WARN


def get_run_info(toml_filepath, num_channels=512):
    """Convert a TOML representation of a ReadFish experiment to conditions that
    can be used used by the analysis function

    Parameters
    ----------
    toml_filepath : str
        Filepath to a configuration TOML file
    num_channels : int
        Total number of channels on the sequencer, expects 512 for MinION and 3000 for
        PromethION

    Returns
    -------
    run_info : dict
        dict with a key per channel, the value maps to an index in `split_conditions`
    split_conditions : list
        List of namedtuples with conditions specified in the TOML file
    reference : str
        The path to the reference MMI file
    caller_settings : dict
        kwargs to pass to the base caller. If not found in the TOML an empty dict
        is returned
    """
    toml_dict = load_config_toml(toml_filepath)

    # Get condition keys, these should be ascending integers
    conditions = [
        k for k in toml_dict["conditions"].keys()
        if isinstance(toml_dict["conditions"].get(k), dict)
    ]

    # If maintain_order, is True: condition keys are sorted -> [0, 1, 2, 3]
    #  else: sorted is used with random.random() to shuffle the keys -> [4, 1, 2, 3]
    #  this sort is applied during the creation of `split_conditions`
    #  If maintain_order is not provided, evaluates True -> sorted
    if toml_dict["conditions"].get("maintain_order", True):
        sort_func = lambda L: sorted(L)
    else:
        sort_func = lambda L: sorted(L, key=lambda k: random())

    # Generate the flowcell lists
    axis = toml_dict["conditions"].get("axis", 1)
    split_channels = generate_flowcell(num_channels, split=len(conditions), axis=axis)

    # convert targets to sets
    for k in conditions:
        cond = toml_dict["conditions"].get(k)
        if not isinstance(cond, dict):
            continue
        cond["coords"] = get_targets(cond["targets"])

        _t = []
        for _k in cond["coords"].keys():
            _t.extend(cond["coords"].get(_k).keys())

        cond["targets"] = set(_t)

    # Create a list of named tuples, these are the conditions
    split_conditions = [
        named_tuple_generator(toml_dict["conditions"].get(k))
        for k in sort_func(toml_dict["conditions"].keys())
        if isinstance(toml_dict["conditions"].get(k), dict)
    ]

    run_info = {
        channel: pos
        for pos, (channels, condition) in enumerate(
            zip(split_channels, split_conditions)
        )
        for channel in channels
    }

    reference = toml_dict["conditions"].get("reference")
    caller_settings = toml_dict.get("caller_settings", {})

    return run_info, split_conditions, reference, caller_settings


def between(pos, coords):
    """Return bool if position is between the coords

    Parameters
    ----------
    pos : int
        Position to check
    coords : tuple
        Region to check between

    Returns
    -------
    bool

    Examples
    --------

    Between can use any valid floats in the `coords` tuple

    >>> between(500, (0, float("inf")))
    True
    >>> between(5, (10, 100))
    False
    >>> any([between(5, (10, 100)), between(245000000, (0, float("inf"))), ])
    True
    >>> any([between(5, (10, 100)), between(-1, (0, float("inf"))), ])
    False

    """
    return min(coords) <= pos <= max(coords)


def setup_logger(name, log_format="%(message)s", log_file=None, level=logging.DEBUG):
    """Setup loggers

    Parameters
    ----------
    name : str
        Name to give the logger
    log_format : str
        logging format string using % formatting
    log_file : str
        File to record logs to, sys.stderr if not set
    level : logging.LEVEL
        Where logging.LEVEL is one of (DEBUG, INFO, WARNING, ERROR, CRITICAL)

    Returns
    -------
    logger
    """
    """Function setup as many loggers as you want"""
    formatter = logging.Formatter(log_format)
    if log_file is not None:
        handler = logging.FileHandler(log_file, mode="w")
    else:
        handler = logging.StreamHandler()
    handler.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.addHandler(handler)

    return logger


if __name__ == "__main__":
    import doctest
    doctest.testmod()



import logging

from read_until import ReadUntilClient


class Client(ReadUntilClient):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Set our extension here
        # TODO: unblock logger
        # TODO: message port?
        # TODO: wait for acquisition
        # Don't bother writing to data dir
        #

    def unblock_read(self, read_channel, read_number, duration=0.1):
        super().unblock_read(
            read_channel=read_channel, read_number=read_number, duration=duration,
        )
        # Log read id of unblocked read



"""
Iter-align.
Design spec.
    1. Grab fastq files from a location.
    2. Align the files ONCE to a reference.
    3. Calculate cumulative coverage information.
    4. Write a list of genomes that are covered at a particular threshold.
    5. Rinse and repeat
"""
import logging
import logging.handlers
import os
import subprocess
import sys
import threading
import time
from io import StringIO

import pandas as pd
import toml
from watchdog.events import FileSystemEventHandler
from watchdog.observers.polling import PollingObserver as Observer

from ru.arguments import get_parser
from ru.utils import nice_join, print_args, send_message, Severity
from read_until_api_v2.load_minknow_rpc import get_rpc_connection, parse_message


DEFAULT_SERVER_HOST = "127.0.0.1"
DEFAULT_LOG_FORMAT = "%(asctime)s - %(name)-20s - %(message)s"
LOG_LEVELS = ("debug", "info", "warning", "error", "critical")
DEFAULT_COVERAGE_DEPTH = 30
DEFAULT_PERCENTAGE_COVERED = 0.99
DEFAULT_CORES = 2


_help = "ReadFish and Run Until, using minimap2"
_cli = (
    (
        "--host",
        dict(
            metavar="HOST",
            help="MinKNOW server host",
            default=DEFAULT_SERVER_HOST,
        ),
    ),
    (
        "--device",
        dict(
            metavar="DEVICE",
            type=str,
            help="Name of the sequencing position e.g. MS29042 or GA10000 etc.",
            required=True,
        ),
    ),
    (
        "--watch",
        dict(
            metavar="FOLDER",
            help="Top Level Folder containing fastq reads.",
            default=None,
        ),
    ),
    (
        "--percent",
        dict(
            metavar="PERCENT",
            help="Default percent of target covered at given depth (default {})".format(DEFAULT_PERCENTAGE_COVERED),
            default=DEFAULT_PERCENTAGE_COVERED,
            type=float,
        ),
    ),
    (
        "--depth",
        dict(
            metavar="DEPTH",
            help="Desired coverage depth (default {})".format(DEFAULT_COVERAGE_DEPTH),
            default=DEFAULT_COVERAGE_DEPTH,
            type=int,
        ),
    ),
    (
        "--threads",
        dict(
            metavar="THREADS",
            help="Set the number of default threads to use for threaded tasks (default {})".format(DEFAULT_CORES),
            default=DEFAULT_CORES,
            type=int,
        ),
    ),
    (
        "--log-level",
        dict(
            metavar="LOG-LEVEL",
            action="store",
            default="info",
            choices=LOG_LEVELS,
            help="One of: {}".format(nice_join(LOG_LEVELS)),
        ),
    ),
    (
        "--log-format",
        dict(
            metavar="LOG-FORMAT",
            action="store",
            default=DEFAULT_LOG_FORMAT,
            help="A standard Python logging format string (default: {!r})".format(
                DEFAULT_LOG_FORMAT.replace("%", "%%")
            ),
        ),
    ),
    (
        "--log-file",
        dict(
            metavar="LOG-FILE",
            action="store",
            default=None,
            help="A filename to write logs to, or None to write to the standard stream (default: None)",
        ),
    ),
    (
        "--toml",
        dict(
            metavar="TOML",
            required=True,
            help="The magic TOML file that will save your life?",
            #type=toml.load,
        ),
    ),
)


def file_dict_of_folder_simple(path, args, logging, fastqdict):
    logger = logging.getLogger("ExistingFileProc")

    file_list_dict = dict()

    counter = 0

    if os.path.isdir(path):

        logger.info("caching existing fastq files in: %s" % (path))

        for path, dirs, files in os.walk(path):

            for f in files:

                counter += 1

                if f.endswith(".fastq") or f.endswith(".fastq.gz"):

                    logger.debug("Processing File {}\r".format(f))
                    filepath = os.path.join(path, f)
                    file_list_dict[filepath] = os.stat(filepath).st_mtime

    logger.info("processed %s files" % (counter))



    logger.info("found %d existing fastq files to process first." % (len(file_list_dict)))

    return file_list_dict


def parse_fastq_file(fastqfilelist,args,logging,masterdf):
    logger = logging.getLogger("ParseFastq")

    with open(os.devnull, 'w') as devnull:
        # Run basic mapping
        minimapcmd = ["minimap2","-ax","map-ont","-t {}".format(args.threads),args.toml['conditions']['reference']] #" ".join(fastqfilelist)]
        minimapcmd.extend(fastqfilelist)
        logger.info(" ".join(minimapcmd))
        minimapoutput = subprocess.Popen(minimapcmd, stdout=subprocess.PIPE,stderr=devnull)
        samcmd = ["samtools","view", "-bS"]
        samoutput = subprocess.Popen(samcmd, stdin=minimapoutput.stdout, stdout=subprocess.PIPE, stderr=devnull)
        #samsortcmd = ["samtools", "sort", "-@2", "-o", "sortedbam.bam"]
        samsortcmd = ["samtools", "sort", "-@{}".format(args.threads)]
        samsortoutput = subprocess.Popen(samsortcmd, stdin=samoutput.stdout, stdout=subprocess.PIPE, stderr=devnull)
        samdepthcmd = ["samtools", "depth", "-a", "/dev/stdin"]
        samdepthoutput = subprocess.Popen(samdepthcmd, stdin=samsortoutput.stdout,stdout=subprocess.PIPE, stderr=devnull, universal_newlines=True)
        minimapoutput.stdout.close()
        samoutput.stdout.close()
        samsortoutput.stdout.close()
        output,err = samdepthoutput.communicate()

        coveragedf = pd.read_csv(StringIO(output),sep="\t",names=['seqid','position','coverage'])


        summarydf = coveragedf.groupby('seqid', as_index=False).agg({'position':'max','coverage':'sum'})


        tempsummarydf = pd.concat([masterdf,summarydf])
        masterdf = tempsummarydf.groupby('seqid', as_index=False).agg({'position': 'max', 'coverage': 'sum'})
        masterdf["tmp"] = masterdf["coverage"]/masterdf['position']
        # Calculate depth at every position
        targets = masterdf[masterdf["tmp"].ge(args.depth)]

        logger.info(targets)

        targets = targets['seqid'].tolist()

        masterdf.drop(['tmp'], axis=1, inplace=True)

        logger.info("Finished processing {}.".format(" ".join(fastqfilelist)))

    return targets,masterdf



def write_new_toml(args,targets):
    for k in args.toml["conditions"].keys():
        curcond = args.toml["conditions"].get(k)
        if isinstance(curcond,dict):

            #newtargets = targets
            #newtargets.extend(curcond["targets"])

            #newtargets = list(dict.fromkeys(newtargets))
            #curcond["targets"]=list(set(newtargets))
            curcond["targets"]=targets

    with open("{}_live".format(args.tomlfile), "w") as f:
        toml.dump(args.toml,f)




class FastqHandler(FileSystemEventHandler):

    def __init__(self, args,logging,messageport,rpc_connection):
        self.args = args
        self.messageport = messageport
        self.connection = rpc_connection
        self.logger = logging.getLogger("FastqHandler")
        self.running = True
        self.fastqdict = dict()
        self.creates = file_dict_of_folder_simple(self.args.watch, self.args, logging,
                                                  self.fastqdict)
        self.t = threading.Thread(target=self.processfiles)

        try:
            self.t.start()
        except KeyboardInterrupt:
            self.t.stop()
            raise


    def processfiles(self):
        self.logger.info("Process Files Inititated")
        self.counter = 0
        self.targets = []
        self.masterdf = pd.DataFrame(columns=['seqid','position','coverage'])

        while self.running:
            currenttime = time.time()
            #for fastqfile, createtime in tqdm(sorted(self.creates.items(), key=lambda x: x[1])):
            fastqfilelist=list()
            for fastqfile, createtime in sorted(self.creates.items(), key=lambda x: x[1]):

                delaytime = 0

                # file created 5 sec ago, so should be complete. For simulations we make the time longer.
                if (int(createtime) + delaytime < time.time()):
                    self.logger.info(fastqfile)
                    del self.creates[fastqfile]
                    self.counter +=1
                    fastqfilelist.append(fastqfile)

                    #print (fastqfile,md5Checksum(fastqfile), "\n\n\n\n")
            targets,self.masterdf = parse_fastq_file(fastqfilelist,self.args,logging,self.masterdf)
            print (targets)
            print (self.targets)
            if len(targets) > len(self.targets):
                updated_targets = set(targets) - set(self.targets)
                update_message = "Updating targets with {}".format(nice_join(updated_targets, conjunction="and"))
                self.logger.info(update_message)
                if not self.args.simulation:
                    send_message(self.connection, update_message, Severity.WARN)
                write_new_toml(self.args,targets)
                self.targets = []
                self.targets = targets.copy()

            if self.masterdf.shape[0] > 0 and self.masterdf.shape[0] == len(self.targets):
                # Every target is covered at the desired coverage level.
                self.logger.info("Every target is covered at at least {}x".format(self.args.depth))
                if not self.args.simulation:
                   self.connection.protocol.stop_protocol()
                   send_message(
                       self.connection,
                       "Iter Align has stopped the run as all targets should be covered by at least {}x".format(
                           self.args.depth
                       ),
                       Severity.WARN,
                   )


            #parse_fastq_file(fastqfile, self.rundict, self.fastqdict, self.args, self.header, self.MinotourConnection)

            #self.args.files_processed += 1

            if currenttime+5 > time.time():
                time.sleep(5)

    def on_created(self, event):
        """Watchdog counts a new file in a folder it is watching as a new file"""
        """This will add a file which is added to the watchfolder to the creates and the info file."""
        # if (event.src_path.endswith(".fastq") or event.src_path.endswith(".fastq.gz")):
        #     self.creates[event.src_path] = time.time()


        # time.sleep(5)
        if (event.src_path.endswith(".fastq") or event.src_path.endswith(".fastq.gz") or event.src_path.endswith(
                ".fq") or event.src_path.endswith(".fq.gz")):
            self.logger.info("Processing file {}".format(event.src_path))
            self.creates[event.src_path] = time.time()

    def on_modified(self, event):
        if (event.src_path.endswith(".fastq") or event.src_path.endswith(".fastq.gz") or event.src_path.endswith(
                ".fq") or event.src_path.endswith(".fq.gz")):
            self.logger.info("Processing file {}".format(event.src_path))
            self.logger.debug("Modified file {}".format(event.src_path))
            self.creates[event.src_path] = time.time()

    def on_moved(self, event):
        if any((event.dest_path.endswith(".fastq"), event.dest_path.endswith(".fastq,gz"),
                event.dest_path.endswith(".gq"), event.dest_path.endswith(".fq.gz"))):
            self.logger.info("Processing file {}".format(event.dest_path))
            self.logger.debug("Modified file {}".format(event.dest_path))
            self.creates[event.dest_path] = time.time()



def main():
    sys.exit(
        "This entry point is deprecated, please use 'readfish align' instead"
    )


def run(parser, args):
    args.tomlfile = args.toml
    args.toml = toml.load(args.toml)
    print (args)

    # TODO: Move logging config to separate configuration file
    # set up logging to file
    logging.basicConfig(level=logging.DEBUG,
                        format='%(levelname)s::%(asctime)s::%(name)s::%(message)s',
                        filename=args.log_file,
                        filemode='w')

    # define a Handler that writes INFO messages or higher to the sys.stderr
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)

    # set a format which is simpler for console use
    formatter = logging.Formatter('%(name)-15s: %(levelname)-8s %(message)s')
    console.setFormatter(formatter)

    # add the handler to the root logger
    logging.getLogger('').addHandler(console)

    # Start by logging sys.argv and the parameters used
    logger = logging.getLogger("Manager")
    logger.info(" ".join(sys.argv))
    print_args(args, logger=logger)

    logger.info("Initialising iterAlign.")

    logger.info("Setting up FastQ monitoring.")



    #### Check if a run is active - if not, wait.

    args.simulation = True
    connection = None
    if args.watch is None:
        args.simulation = False
        logger.info("Creating rpc connection for device {}.".format(args.device))
        try:
            connection, messageport = get_rpc_connection(args.device)
        except ValueError as e:
            print(e)
            sys.exit(1)

        send_message(connection, "Iteralign Connected to MinKNOW", Severity.WARN)

        logger.info("Loaded RPC")
        while parse_message(connection.acquisition.current_status())['status'] != "PROCESSING":
            time.sleep(1)
        #### Check if we know where data is being written to , if not... wait
        args.watch = parse_message(connection.acquisition.get_acquisition_info())['config_summary']['reads_directory']

    else:
        messageport = ""


    event_handler = FastqHandler(args,logging,messageport,connection)
    # This block handles the fastq
    observer = Observer()
    observer.schedule(event_handler, path=args.watch, recursive=True)
    observer.daemon = True




    try:

        observer.start()
        logger.info("FastQ Monitoring Running.")
        while 1:
            time.sleep(1)

    except KeyboardInterrupt:

        logger.info("Exiting - Will take a few seconds to clean up!")

        observer.stop()
        observer.join()

        os._exit(0)





"""unblock_all.py

ReadUntil implementation that will only unblock reads. This should result in
a read length histogram that has very short peaks (~280-580bp) as these are the
smallest chunks that we can acquire. If you are not seeing these peaks, the
`split_reads_after_seconds` parameter in the configuration file may need to be
edited to 0.2-0.4:
(<MinKNOW_folder>/ont-python/lib/python2.7/site-packages/bream4/configuration)
"""
# Core imports
import concurrent.futures
import functools
import logging
import sys
import time
from timeit import default_timer as timer
import traceback
from multiprocessing import TimeoutError
from multiprocessing.pool import ThreadPool
from collections import defaultdict, deque, Counter

# Pypi imports
import toml
import numpy as np

# Read Until imports
import read_until
from read_until.arguments import get_parser
from read_until.utils import print_args


class ThreadPoolExecutorStackTraced(concurrent.futures.ThreadPoolExecutor):
    """ThreadPoolExecutor records only the text of an exception,
    this class will give back a bit more."""

    def submit(self, fn, *args, **kwargs):
        """Submits the wrapped function instead of `fn`"""
        return super(ThreadPoolExecutorStackTraced, self).submit(
            self._function_wrapper, fn, *args, **kwargs
        )

    def _function_wrapper(self, fn, *args, **kwargs):
        """Wraps `fn` in order to preserve the traceback of any kind of
        raised exception

        """
        try:
            return fn(*args, **kwargs)
        except Exception:
            raise sys.exc_info()[0](traceback.format_exc())


def setup_logger(name, log_file, level=logging.DEBUG):
    """Function setup as many loggers as you want"""
    formatter = logging.Formatter("%(message)s")
    handler = logging.FileHandler(log_file, mode="w")
    handler.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.setLevel(level)
    logger.addHandler(handler)

    return logger


from uuid import uuid4
from fast5_research import Fast5
def rpc_to_fast5(channel, read, dtype):
    """
    ['ByteSize', 'Clear', 'ClearExtension', 'ClearField', 'CopyFrom', 'DESCRIPTOR',
     'DiscardUnknownFields', 'Extensions', 'FindInitializationErrors', 'FromString',
     'HasExtension', 'HasField', 'IsInitialized', 'ListFields', 'MergeFrom',
     'MergeFromString', 'ParseFromString', 'RegisterExtension', 'SerializePartialToString',
     'SerializeToString', 'SetInParent', 'UnknownFields', 'WhichOneof',
     '_CheckCalledFromGeneratedFile', '_SetListener', '__class__', '__deepcopy__',
     '__delattr__', '__dir__', '__doc__', '__eq__', '__format__', '__ge__', '__getattribute__',
     '__getstate__', '__gt__', '__hash__', '__init__', '__init_subclass__', '__le__', '__lt__',
     '__module__', '__ne__', '__new__', '__reduce__', '__reduce_ex__', '__repr__', '__setattr__',
     '__setstate__', '__sizeof__', '__slots__', '__str__', '__subclasshook__', '__unicode__',
     '_extensions_by_name', '_extensions_by_number', 'chunk_classifications', 'chunk_length',
     'chunk_start_sample', 'id', 'median', 'median_before', 'number', 'raw_data', 'start_sample']

    Parameters
    ----------
    channel
    read

    Returns
    -------

    """
    filename = f"{read.id}.fast5"
    # mean, stdv, n = 40.0, 2.0, 10000
    # raw_data = np.random.laplace(mean, stdv / np.sqrt(2), int(n))
    raw_data = np.frombuffer(read.raw_data, dtype=dtype)

    # example of how to digitize data
    start, stop = int(min(raw_data - 1)), int(max(raw_data + 1))
    rng = stop - start
    digitisation = 8192.0
    bins = np.arange(start, stop, rng / digitisation)

    # np.int16 is required, the library will refuse to write anything other
    raw_data = np.digitize(raw_data, bins).astype(np.int16)

    # The following are required meta data
    channel_id = {
        'digitisation': digitisation,
        'offset': 0,
        'range': rng,
        'sampling_rate': 4000,
        'channel_number': channel,
    }
    read_id = {
        'start_time': 0,
        'duration': len(raw_data),
        'read_number': read.number,
        'start_mux': 1,
        'read_id': str(uuid4()),
        'scaling_used': 1,
        'median_before': read.median_before,
    }
    tracking_id = {
        'exp_start_time': '1970-01-01T00:00:00Z',
        'run_id': str(uuid4()).replace('-', ''),
        'flow_cell_id': 'FAH00000',
    }
    context_tags = {}

    with Fast5.New(filename, 'w', tracking_id=tracking_id, context_tags=context_tags, channel_id=channel_id) as h:
        h.set_raw(raw_data, meta=read_id, read_number=1)
    return True


def simple_analysis(
    client, batch_size=512, throttle=0.1, unblock_duration=0.5, chunk_log=None, paf_log=None, toml_path=None, log_level="INFO"
):
    """Analysis function

    Parameters
    ----------
    client : read_until.ReadUntilClient
        An instance of the ReadUntilClient object
    batch_size : int
        The number of reads to be retrieved from the ReadUntilClient at a time
    throttle : int or float
        The number of seconds interval between requests to the ReadUntilClient
    unblock_duration : int or float
        Time, in seconds, to apply unblock voltage
    chunk_log : str
        Log file to log chunk data to
    paf_log : str
        Log file to log paf data to
    log_level: str
        Default log level

    Returns
    -------
    None
    """
    display = True
    t = toml.load(toml_path)
    logger = logging.getLogger(__name__)

    # DefaultDict[int: collections.deque[Tuple[str, ndarray]]]
    #  tuple is (read_id, previous_signal)
    # TODO: tuple could use read_number instead
    previous_signal = defaultdict(functools.partial(deque, maxlen=1))
    # count how often a read is seen
    tracker = defaultdict(Counter)
    # decided
    decided_reads = {}


    cl = setup_logger("DEC", chunk_log,level=logging.INFO)
    pf = setup_logger("PAF", paf_log,level=logging.INFO)
    l_string = (
        "client_iteration",
        "read_in_loop",
        "read_id",
        "channel",
        "read_number",
        "raw_len",
        "estimated_bases",
        "start_analysis",
        "end_analysis",
        "timestamp",
    )
    cl.debug("\t".join(l_string))
    l_string = "\t".join(("{}" for _ in l_string))

    tracker = defaultdict(Counter)

    longestreads = dict()

    loop_counter = 0
    r = 0
    readcounter = 0
    while client.is_running:
        logger.info(f"Client is running {loop_counter}")
        loop_counter += 1
        t0 = timer()

        for r, (channel, read) in enumerate(
                client.get_read_chunks(batch_size=batch_size, last=True),
                start=1
        ):
            logger.info("got reads")
            read_start_time = timer()
            if read.number not in tracker[channel]:
                tracker[channel].clear()
            tracker[channel][read.number] += 1

            old_read_id, old_signal = previous_signal.get(channel, (("", np.empty(0, dtype=client.signal_dtype)),))[0]
            x = np.frombuffer(read.raw_data, dtype=client.signal_dtype)
            logger.info(client.signal_dtype)
            logger.info("{} {}".format(type(old_signal), len(old_signal)))
            logger.info("{} {}".format(type(x), len(x)))
            if old_read_id == read.id:
                signal = np.concatenate((old_signal, x,))
            else:
                signal = x

            previous_signal[channel].append((read.id, signal))

            cl.debug(
                l_string.format(
                    loop_counter,
                    r,
                    read.id,
                    channel,
                    read.number,
                    len(signal),
                    len(signal) / 450,
                    read_start_time,
                    timer(),
                    time.time(),
                    )
            )
            # print(dir(read))
            if not rpc_to_fast5(channel, read, signal.dtype):
                logger.info("💣")
            break
        # break

        t1 = timer()
        # logger.info("Took {} for {} reads".format(t1 - t0, r))
        # limit the rate at which we make requests
        if t0 + throttle > t1:
            time.sleep(throttle + t0 - t1)
    else:
        logger.info("Finished analysis of reads as client stopped.")


def run_workflow(client, analysis_worker, n_workers, run_time, runner_kwargs=None):
    """Run an analysis function against a ReadUntilClient

    Parameters
    ----------
    client : read_until.ReadUntilClient
        An instance of the ReadUntilClient object
    analysis_worker : partial function
        Analysis function to process reads, should exit when client.is_running == False
    n_workers : int
        Number of analysis worker functions to run
    run_time : int
        Time, in seconds, to run the analysis for
    runner_kwargs : dict
        Keyword arguments to pass to client.run()

    Returns
    -------
    list
        Results from the analysis function, one item per worker

    """
    if runner_kwargs is None:
        runner_kwargs = dict()

    logger = logging.getLogger("Manager")

    results = []
    pool = ThreadPool(n_workers)
    logger.info("Creating {} workers".format(n_workers))
    try:
        # start the client
        client.run(**runner_kwargs)
        # start a pool of workers
        for _ in range(n_workers):
            results.append(pool.apply_async(analysis_worker))
        pool.close()
        # wait a bit before closing down
        time.sleep(run_time)
        logger.info("Sending reset")
        client.reset()
        pool.join()
    except KeyboardInterrupt:
        logger.info("Caught ctrl-c, terminating workflow.")
        client.reset()

    # collect results (if any)
    collected = []
    for result in results:
        try:
            res = result.get(5)
        except TimeoutError:
            logger.warning("Worker function did not exit successfully.")
            # collected.append(None)
        except Exception as e:
            logger.exception("EXCEPT", exc_info=e)
            # logger.warning("Worker raise exception: {}".format(repr(e)))
        else:
            logger.info("Worker exited successfully.")
            collected.append(res)
    pool.terminate()
    return collected


def main():
    extra_args = (
        (
            "--chunk-log",
            dict(
                help="Chunk log",
                default="chunk_log.log",
            )
        ),
        (
            "--paf-log",
            dict(
                help="PAF log file",
                default="paf_output_log.paf",
            )
        ),
        (
            "--toml",
            dict(
                metavar="TOML",
                required=True,
                help="TOML file specifying experimental parameters. Here we only "
                     "use `guppy_connection` and the `reference` location.",
            ),
        ),
    )
    parser, args = get_parser(extra_args=extra_args, file=__file__)

    print (args)


    # TODO: Move logging config to separate configuration file
    # set up logging to file for DEBUG messages and above
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(levelname)s %(asctime)s %(name)s %(message)s",
        filename=args.log_file,
        filemode="w",
    )

    # define a Handler that writes INFO messages or higher to the sys.stderr
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)

    # set a format which is simpler for console use
    formatter = logging.Formatter(args.log_format)
    console.setFormatter(formatter)

    # add the handler to the root logger
    logging.getLogger("").addHandler(console)

    # Start by logging sys.argv and the parameters used
    logger = logging.getLogger("Manager")
    logger.info(" ".join(sys.argv))
    print_args(args, logger=logger)

    read_until_client = read_until.ReadUntilClient(
        mk_host=args.host,
        device=args.device,
        # one_chunk=args.one_chunk,
        filter_strands=True,
        cache_type=args.read_cache,
        cache_size=args.cache_size,
    )

    analysis_worker = functools.partial(
        simple_analysis,
        read_until_client,
        unblock_duration=args.unblock_duration,
        throttle=args.throttle,
        chunk_log=args.chunk_log,
        paf_log=args.paf_log,
        toml_path=args.toml,
        log_level=args.log_level,
    )

    results = run_workflow(
        read_until_client,
        analysis_worker,
        args.workers,
        args.run_time,
        runner_kwargs={
            # "min_chunk_size": args.min_chunk_size,
            "first_channel": args.channels[0],
            "last_channel": args.channels[-1],
        },
    )
    # No results returned


if __name__ == "__main__":
    main()



import argparse
import importlib

from ._version import __version__


def main():
    parser = argparse.ArgumentParser(
        prog="readfish",
        epilog="See '<command> --help' to read about a specific sub-command.",
    )
    version = "readfish {}".format(__version__)
    parser.add_argument("--version", action="version", version=version)
    subparsers = parser.add_subparsers(dest="command", help="Sub-commands")

    cmds = [
        ("targets", "ru_gen"),
        ("align", "iteralign"),
        ("centrifuge", "iteralign_centrifuge"),
        ("unblock-all", "unblock_all"),
        ("validate", "validate"),
        ("summary", "summarise_fq")
    ]
    for cmd, module in cmds:
        _module = importlib.import_module("ru.{}".format(module))
        _parser = subparsers.add_parser(cmd, help=_module._help)
        for *flags, opts in _module._cli:
            _parser.add_argument(*flags, **opts)
        _parser.set_defaults(func=_module.run)

    args = parser.parse_args()
    if args.command is not None:
        args.func(parser, args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



import gzip
from pathlib import Path
from statistics import mean, median, stdev
from collections import defaultdict
import sys

import toml
import mappy as mp


_help = "Summary stats from FASTQ files"
_cli = (
    ("toml", dict(help="TOML configuration file"),),
    ("dir", dict(help="Called files from the ReadFish experiment"),),
)


def readfq(fp): # this is a generator function
    """Read FASTA/Q records from file handle
    https://github.com/lh3/readfq/blob/091bc699beee3013491268890cc3a7cbf995435b/readfq.py
    """
    last = None # this is a buffer keeping the last unprocessed line
    while True: # mimic closure; is it a bad idea?
        if not last: # the first record or a record following a fastq
            for l in fp: # search for the start of the next record
                if l[0] in '>@': # fasta/q header line
                    last = l[:-1] # save this line
                    break
        if not last: break
        name, seqs, last = last[1:].partition(" ")[0], [], None
        for l in fp: # read the sequence
            if l[0] in '@+>':
                last = l[:-1]
                break
            seqs.append(l[:-1])
        if not last or last[0] != '+': # this is a fasta record
            yield name, ''.join(seqs), None # yield a fasta record
            if not last: break
        else: # this is a fastq record
            seq, leng, seqs = ''.join(seqs), 0, []
            for l in fp: # read the quality
                seqs.append(l[:-1])
                leng += len(l) - 1
                if leng >= len(seq): # have read enough quality
                    last = None
                    yield name, seq, ''.join(seqs); # yield a fastq record
                    break
            if last: # reach EOF before reading enough quality
                yield name, seq, None # yield a fasta record instead
                break


def get_fq(directory):
    types = ([".fastq"], [".fastq", ".gz"], [".fq"], [".fq", ".gz"])
    files = (str(p.resolve()) for p in Path(directory).glob("**/*") if p.suffixes in types)
    yield from files


def icumsum(arr):
    total = 0
    for i, x in enumerate(arr):
        total += x
        yield i, total


def N50(arr):
    if not isinstance(arr, list):
        arr = list(arr)
    arr.sort()
    s = sum(arr)
    return int(arr[[i for i, c in icumsum(arr) if c >= s * 0.5][0]])


def main():
    sys.exit(
       "This entry point is deprecated, please use 'readfish summary' instead"
    )


def run(parser, args):
    t = toml.load(args.toml)
    reference = t["conditions"].get("reference", "")

    if not Path(reference).is_file():
        raise FileNotFoundError("reference file not found at: {}".format(reference))

    mapper = mp.Aligner(reference, preset="map-ont")

    print("Using reference: {}".format(reference), file=sys.stderr)

    pre_res = defaultdict(list)

    for f in get_fq(args.dir):
        if f.endswith(".gz"):
            fopen = gzip.open
        else:
            fopen = open

        with fopen(f, "rt") as fh:
            for name, seq, _ in readfq(fh):
                # Map seq, only use first mapping (a bit janky)
                for r in mapper.map(seq):
                    pre_res[r.ctg].append(len(seq))
                    break

    header = [
        "contig",
        "number",
        "sum",
        "min",
        "max",
        "std",
        "mean",
        "median",
        "N50",
    ]
    res = [header]
    for ctg, data in sorted(pre_res.items()):
        if len(data) < 2:
            print("Skipping contig {}, too few mappings".format(ctg), file=sys.stderr)
            continue
        res.append(
            [
                ctg,
                "{:.0f}".format(len(data)),
                "{:.0f}".format(sum(data)),
                "{:.0f}".format(min(data)),
                "{:.0f}".format(max(data)),
                "{:.0f}".format(stdev(data)),
                "{:.0f}".format(mean(data)),
                "{:.0f}".format(median(data)),
                "{:.0f}".format(N50(data)),
            ]
        )

    m = list(max(map(len, x)) + 2 for x in list(map(list, zip(*res))))
    m[0] -= 2
    s = "{:>{}}" * len(m)
    for x in res:
        c = x + m
        c[::2] = x
        c[1::2] = m
        print(s.format(*c), file=sys.stderr)


if __name__ == "__main__":
    main()



"""basecall.py

Extension of pyguppy Caller that maintains a connection to the basecaller

"""
import logging

import mappy as mp
import numpy as np

from pyguppyclient.client import GuppyBasecallerClient
from pyguppyclient.decode import ReadData as GuppyRead


__all__ = ["GuppyCaller"]

logger = logging.getLogger("RU_basecaller")


def _create_guppy_read(reads, signal_dtype, previous_signal):
    """Convert a read from MinKNOW RPC to GuppyRead

    Parameters
    ----------
    reads : List[Tuple[int, minknow.rpc.read]]
        List of Tuple, containing (channel, read)
    signal_dtype : np.dtype
        A dtype that can be used by numpy to convert the raw data
    previous_signal : dict
        Dict containing previous signal segments

    Yields
    ------
    channel : int
    read_number : int
    GuppyRead
    """
    for read_id, channel, read_number, signal in _concat_signal(
        reads, signal_dtype, previous_signal
    ):
        read_obj = GuppyRead(signal, read_id, 0, 1)
        previous_signal[channel].append((read_id, read_obj.signal))
        yield channel, read_number, read_obj


def _concat_signal(reads, signal_dtype, previous_signal):
    for channel, read in reads:
        old_read_id, old_signal = previous_signal.get(
            channel, (("", np.empty(0, dtype=signal_dtype)),)
        )[0]

        if old_read_id == read.id:
            signal = np.concatenate(
                (old_signal, np.frombuffer(read.raw_data, dtype=signal_dtype))
            )
        else:
            signal = np.frombuffer(read.raw_data, dtype=signal_dtype)

        yield read.id, channel, read.number, signal


class GuppyCaller(GuppyBasecallerClient):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.connect()

    def basecall_minknow(self, reads, signal_dtype, prev_signal, decided_reads):
        """Guppy basecaller wrapper for MinKNOW RPC reads

        Parameters
        ----------
        reads : iterable[Tuple[int, rpc.Read]]
            List or generator of tuples containing (channel, MinKNOW.rpc.Read)
        signal_dtype
            Numpy dtype of the raw data
        prev_signal : DefaultDict[int: collections.deque[Tuple[str, np.ndarray]]]
            Dictionary of previous signal fragment from a channel
        decided_reads : Dict[int: str]
            Dictionary of channels with the last read id a decision was made for

        Yields
        ------
        read_info : tuple
            Tuple of read info (channel, read_number)
        read_id : str
        sequence : str
        sequence_length : int
        quality : str
        """
        done = 0
        read_counter = 0

        hold = {}
        for channel, read_number, read in _create_guppy_read(
            reads, signal_dtype, prev_signal
        ):
            if read.read_id == decided_reads.get(channel, ""):
                continue

            hold[read.read_id] = (channel, read_number)
            try:
                self.pass_read(read)
            except Exception as e:
                logger.warning("Skipping read: {} due to {}".format(read.read_id, e))
                hold.pop(read.read_id)
                continue
            read_counter += 1

        while done < read_counter:
            res = self._get_called_read()

            if res is None:
                continue

            read, called = res

            yield hold.pop(
                read.read_id
            ), read.read_id, called.seq, called.seqlen, called.qual
            done += 1


class Mapper:
    def __init__(self, index):
        self.index = index
        if self.index:
            self.mapper = mp.Aligner(self.index, preset="map-ont")
            self.initialised = True
        else:
            self.mapper = None
            self.initialised = False

    def map_read(self, seq):
        return self.mapper.map(seq)

    def map_reads(self, calls):
        for read_id, seq in calls:
            yield read_id, list(self.mapper.map(seq))

    def map_reads_2(self, calls):
        """Align reads against a reference

        Parameters
        ----------
        calls : iterable [tuple,  str, str, int, str]
            An iterable of called reads from PerpetualCaller.basecall_minknow

        Yields
        ------
        read_info : tuple
            Tuple of read info (channel, read_number)
        read_id : str
        sequence : str
        sequence_length : int
        mapping_results : list
        """
        for read_info, read_id, seq, seq_len, quality in calls:
            yield read_info, read_id, seq_len, list(self.mapper.map(seq))



"""unblock_all.py

ReadUntil implementation that will only unblock reads. This should result in
a read length histogram that has very short peaks (~280-580bp) as these are the
smallest chunks that we can acquire. If you are not seeing these peaks, the
`split_reads_after_seconds` parameter in the configuration file may need to be
edited to 0.2-0.4:
(<MinKNOW_folder>/ont-python/lib/python2.7/site-packages/bream4/configuration)
"""
# Core imports
import concurrent.futures
import functools
import logging
import sys
import time
import traceback
from collections import defaultdict, deque, Counter
from multiprocessing import TimeoutError
from multiprocessing.pool import ThreadPool
from pathlib import Path
from timeit import default_timer as timer

# Third party imports
import read_until_api_v2 as read_until
import toml

from ru.arguments import get_parser, BASE_ARGS
from ru.basecall import Mapper as CustomMapper
from ru.basecall import GuppyCaller as Caller
from ru.utils import print_args, get_run_info, between, setup_logger, describe_experiment
from ru.utils import send_message, Severity


_help = "Run targeted sequencing"
_cli = BASE_ARGS + (
    (
        "--toml",
        dict(
            metavar="TOML",
            required=True,
            help="TOML file specifying experimental parameters",
        ),
    ),
    (
        "--paf-log",
        dict(
            help="PAF log",
            default="paflog.log",
        )
    ),
    (
        "--chunk-log",
        dict(
            help="Chunk log",
            default="chunk_log.log",
        )
    ),
)

class ThreadPoolExecutorStackTraced(concurrent.futures.ThreadPoolExecutor):
    """ThreadPoolExecutor records only the text of an exception,
    this class will give back a bit more."""

    def submit(self, fn, *args, **kwargs):
        """Submits the wrapped function instead of `fn`"""
        return super(ThreadPoolExecutorStackTraced, self).submit(
            self._function_wrapper, fn, *args, **kwargs
        )

    def _function_wrapper(self, fn, *args, **kwargs):
        """Wraps `fn` in order to preserve the traceback of any kind of
        raised exception

        """
        try:
            return fn(*args, **kwargs)
        except Exception:
            raise sys.exc_info()[0](traceback.format_exc())


def simple_analysis(
        client,
        batch_size=512,
        throttle=0.1,
        unblock_duration=0.5,
        cl=None,
        pf=None,
        live_toml_path=None,
        flowcell_size=512,
        dry_run=False,
        run_info=None,
        conditions=None,
        mapper=None,
        caller_kwargs=None,
):
    """Analysis function

    Parameters
    ----------
    client : read_until.ReadUntilClient
        An instance of the ReadUntilClient object
    batch_size : int
        The number of reads to be retrieved from the ReadUntilClient at a time
    throttle : int or float
        The number of seconds interval between requests to the ReadUntilClient
    unblock_duration : int or float
        Time, in seconds, to apply unblock voltage
    cl : logging.Logger
        Log file to log chunk data to
    pf : logging.Logger
        Log file to log alignments to
    live_toml_path : str
        Path to a `live` TOML configuration file for ReadFish. If this exists when
        the run starts it will be deleted
    flowcell_size : int
        The number of channels on the flowcell, 512 for MinION and 3000 for PromethION
    dry_run : bool
        If True unblocks are replaced with `stop_receiving` commands
    run_info : dict
        Dictionary of {channel: index} where index corresponds to an index in `conditions`
    conditions : list
        Experimental conditions as List of namedtuples.
    mapper : mappy.Aligner
    caller_kwargs : dict

    Returns
    -------
    None
    """
    # Init logger for this function
    logger = logging.getLogger(__name__)

    # Delete live TOML file if it exists
    live_toml_path = Path(live_toml_path)
    if live_toml_path.is_file():
        live_toml_path.unlink()

    # TODO: test this
    # Write channels.toml
    d = {"conditions": {str(v): {"channels": [], "name": conditions[v].name} for k, v in run_info.items()}}
    for k, v in run_info.items():
        d["conditions"][str(v)]["channels"].append(k)

    channels_out = str(client.mk_run_dir / "channels.toml")
    with open(channels_out, "w") as fh:
        fh.write("# This file is written as a record of the condition each channel is assigned.\n")
        fh.write("# It may be changed or overwritten if you restart ReadFish.\n")
        fh.write("# In the future this file may become a CSV file.\n")
        toml.dump(d, fh)

    caller = Caller(**caller_kwargs)
    # What if there is no reference or an empty MMI

    # DefaultDict[int: collections.deque[Tuple[str, ndarray]]]
    #  tuple is (read_id, previous_signal)
    # TODO: tuple should use read_number instead
    previous_signal = defaultdict(functools.partial(deque, maxlen=1))
    # count how often a read is seen
    tracker = defaultdict(Counter)
    # decided
    decided_reads = {}
    strand_converter = {1: "+", -1: "-"}

    read_id = ""

    # TODO: partial-ise / lambda unblock to take the unblock duration
    if dry_run:
        decision_dict = {
            "stop_receiving": client.stop_receiving_read,
            "proceed": None,
            "unblock": client.stop_receiving_read,
        }
        send_message(client.connection,"This is a test run. No unblocks will occur.",Severity.WARN)
    else:
        decision_dict = {
            "stop_receiving": client.stop_receiving_read,
            "proceed": None,
            "unblock": lambda c, n: client.unblock_read(c, n, unblock_duration, read_id),
        }
        send_message(client.connection, "This is a live run. Unblocks will occur.", Severity.WARN)
    decision_str = ""
    below_threshold = False
    exceeded_threshold = False

    l_string = (
        "client_iteration",
        "read_in_loop",
        "read_id",
        "channel",
        "read_number",
        "seq_len",
        "counter",
        "mode",
        "decision",
        "condition",
        "min_threshold",
        "count_threshold",
        "start_analysis",
        "end_analysis",
        "timestamp",
    )
    cl.debug("\t".join(l_string))
    l_string = "\t".join(("{}" for _ in l_string))
    loop_counter = 0
    while client.is_running:
        if live_toml_path.is_file():
            # Reload the TOML config from the *_live file
            run_info, conditions, new_reference, _ = get_run_info(live_toml_path, flowcell_size)

            # Check the reference path if different from the loaded mapper
            if new_reference != mapper.index:
                old_reference = mapper.index
                # Log to file and MinKNOW interface
                logger.info("Reloading mapper")
                send_message(client.connection, "Reloading mapper. ReadFish paused.", Severity.INFO)

                # Update mapper client.
                mapper = CustomMapper(new_reference)
                # Log on success
                logger.info("Reloaded mapper")

                # If we've reloaded a reference, delete the previous one
                if old_reference:
                    logger.info("Deleting old mmi {}".format(old_reference))
                    # We now delete the old mmi file.
                    Path(old_reference).unlink()
                    logger.info("Old mmi deleted.")

        # TODO: Fix the logging to just one of the two in use

        if not mapper.initialised:
            time.sleep(throttle)
            continue

        loop_counter += 1
        t0 = timer()
        r = 0

        for read_info, read_id, seq_len, results in mapper.map_reads_2(
                caller.basecall_minknow(
                    reads=client.get_read_chunks(batch_size=batch_size, last=True),
                    signal_dtype=client.signal_dtype,
                    prev_signal=previous_signal,
                    decided_reads=decided_reads,
                )
        ):
            r += 1
            read_start_time = timer()
            channel, read_number = read_info
            if read_number not in tracker[channel]:
                tracker[channel].clear()
            tracker[channel][read_number] += 1

            mode = ""
            exceeded_threshold = False
            below_threshold = False

            log_decision = lambda: cl.debug(
                l_string.format(
                    loop_counter,
                    r,
                    read_id,
                    channel,
                    read_number,
                    seq_len,
                    tracker[channel][read_number],
                    mode,
                    getattr(conditions[run_info[channel]], mode, mode),
                    conditions[run_info[channel]].name,
                    below_threshold,
                    exceeded_threshold,
                    read_start_time,
                    timer(),
                    time.time(),
                )
            )

            # Control channels
            if conditions[run_info[channel]].control:
                mode = "control"
                log_decision()
                client.stop_receiving_read(channel, read_number)
                continue

            # This is an analysis channel
            # Below minimum chunks
            if tracker[channel][read_number] <= conditions[run_info[channel]].min_chunks:
                below_threshold = True

            # Greater than or equal to maximum chunks
            if tracker[channel][read_number] >= conditions[run_info[channel]].max_chunks:
                exceeded_threshold = True

            # No mappings
            if not results:
                mode = "no_map"

            hits = set()
            for result in results:
                pf.debug("{}\t{}\t{}".format(read_id, seq_len, result))
                hits.add(result.ctg)

            if hits & conditions[run_info[channel]].targets:
                # Mappings and targets overlap
                coord_match = any(
                    between(r.r_st, c)
                    for r in results
                    for c in conditions[run_info[channel]]
                        .coords.get(strand_converter.get(r.strand), {})
                        .get(r.ctg, [])
                )
                if len(hits) == 1:
                    if coord_match:
                        # Single match that is within coordinate range
                        mode = "single_on"
                    else:
                        # Single match to a target outside coordinate range
                        mode = "single_off"
                elif len(hits) > 1:
                    if coord_match:
                        # Multiple matches with at least one in the correct region
                        mode = "multi_on"
                    else:
                        # Multiple matches to targets outside the coordinate range
                        mode = "multi_off"

            else:
                # No matches in mappings
                if len(hits) > 1:
                    # More than one, off-target, mapping
                    mode = "multi_off"
                elif len(hits) == 1:
                    # Single off-target mapping
                    mode = "single_off"

            # This is where we make our decision:
            # Get the associated action for this condition
            decision_str = getattr(conditions[run_info[channel]], mode)
            # decision is an alias for the functions "unblock" or "stop_receiving"
            decision = decision_dict[decision_str]

            # If max_chunks has been exceeded AND we don't want to keep sequencing we unblock
            if exceeded_threshold and decision_str != "stop_receiving":
                mode = "exceeded_max_chunks_unblocked"
                client.unblock_read(channel, read_number, unblock_duration, read_id)

            # TODO: WHAT IS GOING ON?!
            #  I think that this needs to change between enrichment and depletion
            # If under min_chunks AND any mapping mode seen we unblock
            # if below_threshold and mode in {"single_off", "multi_off"}:
            if below_threshold and mode in {
                "single_on",
                "single_off",
                "multi_on",
                "multi_off",
            }:
                mode = "below_min_chunks_unblocked"
                client.unblock_read(channel, read_number, unblock_duration, read_id)

            # proceed returns None, so we send no decision; otherwise unblock or stop_receiving
            elif decision is not None:
                decided_reads[channel] = read_id
                decision(channel, read_number)

            log_decision()

        t1 = timer()
        if r > 0:
            s1 = "{}R/{:.5f}s"
            logger.info(s1.format(r, t1 - t0))
        # limit the rate at which we make requests
        if t0 + throttle > t1:
            time.sleep(throttle + t0 - t1)
    else:
        send_message(client.connection, "ReadFish Client Stopped.", Severity.WARN)
        caller.disconnect()
        logger.info("Finished analysis of reads as client stopped.")


def run_workflow(client, analysis_worker, n_workers, run_time, runner_kwargs=None):
    """Run an analysis function against a ReadUntilClient

    Parameters
    ----------
    client : read_until.ReadUntilClient
        An instance of the ReadUntilClient object
    analysis_worker : partial function
        Analysis function to process reads, should exit when client.is_running == False
    n_workers : int
        Number of analysis worker functions to run
    run_time : int
        Time, in seconds, to run the analysis for
    runner_kwargs : dict
        Keyword arguments to pass to client.run()

    Returns
    -------
    list
        Results from the analysis function, one item per worker

    """
    if runner_kwargs is None:
        runner_kwargs = dict()

    logger = logging.getLogger("Manager")

    results = []
    pool = ThreadPool(n_workers)
    logger.info("Creating {} workers".format(n_workers))
    try:
        # start the client
        client.run(**runner_kwargs)
        # start a pool of workers
        for _ in range(n_workers):
            results.append(pool.apply_async(analysis_worker))
        pool.close()
        # wait a bit before closing down
        time.sleep(run_time)
        logger.info("Sending reset")
        client.reset()
        pool.join()
    except KeyboardInterrupt:
        logger.info("Caught ctrl-c, terminating workflow.")
        client.reset()
    except Exception as e:
        logger.exception("Got exception", exc_info=e)
        client.reset()
        raise

    # collect results (if any)
    collected = []
    for result in results:
        try:
            res = result.get(3)
        except TimeoutError:
            logger.warning("Worker function did not exit successfully.")
            # collected.append(None)
        except Exception as e:
            logger.exception("EXCEPT", exc_info=e)
            # logger.warning("Worker raise exception: {}".format(repr(e)))
        else:
            logger.info("Worker exited successfully.")
            collected.append(res)
    pool.terminate()
    return collected


def main():
    sys.exit(
        "This entry point is deprecated, please use 'readfish targets' instead"
    )


def run(parser, args):
    # set up logging to file for DEBUG messages and above
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s %(name)s %(message)s",
        filename=args.log_file,
        filemode="w",
    )

    # define a Handler that writes INFO messages or higher to the sys.stderr
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)

    # set a format which is simpler for console use
    formatter = logging.Formatter(args.log_format)
    console.setFormatter(formatter)

    # add the handler to the root logger
    logging.getLogger("").addHandler(console)

    # Start by logging sys.argv and the parameters used
    logger = logging.getLogger("Manager")
    logger.info(" ".join(sys.argv))
    print_args(args, logger=logger)

    # Setup chunk and paf logs
    chunk_logger = setup_logger("DEC", log_file=args.chunk_log)
    paf_logger = setup_logger("PAF", log_file=args.paf_log)

    # Parse configuration TOML
    # TODO: num_channels is not configurable here, should be inferred from client
    run_info, conditions, reference, caller_kwargs = get_run_info(args.toml, num_channels=512)
    live_toml = Path("{}_live".format(args.toml))

    # Load Minimap2 index
    logger.info("Initialising minimap2 mapper")
    mapper = CustomMapper(reference)
    logger.info("Mapper initialised")

    read_until_client = read_until.ReadUntilClient(
        mk_host=args.host,
        mk_port=args.port,
        device=args.device,
        # one_chunk=args.one_chunk,
        filter_strands=True,
        # TODO: test cache_type by passing a function here
        cache_type=args.read_cache,
        cache_size=args.cache_size,
    )

    send_message(
        read_until_client.connection,
        "ReadFish is controlling sequencing on this device. You use it at your own risk.",
        Severity.WARN,
    )

    for message, sev in describe_experiment(conditions, mapper):
        logger.info(message)

        send_message(
            read_until_client.connection,
            message,
            sev,
        )

    """
    This experiment has N regions on the flowcell.

    using reference: /path/to/ref.mmi

    Region i:NAME (control=bool) has X targets of which Y are found in the reference.
    reads will be unblocked when [u,v], sequenced when [w,x] and polled for more data when [y,z].
    """

    # FIXME: currently flowcell size is not included, this should be pulled from
    #  the read_until_client
    analysis_worker = functools.partial(
        simple_analysis,
        read_until_client,
        unblock_duration=args.unblock_duration,
        throttle=args.throttle,
        batch_size=args.batch_size,
        cl=chunk_logger,
        pf=paf_logger,
        live_toml_path=live_toml,
        dry_run=args.dry_run,
        run_info=run_info,
        conditions=conditions,
        mapper=mapper,
        caller_kwargs=caller_kwargs,
    )

    results = run_workflow(
        read_until_client,
        analysis_worker,
        args.workers,
        args.run_time,
        runner_kwargs={
            # "min_chunk_size": args.min_chunk_size,
            "first_channel": min(args.channels),
            "last_channel": max(args.channels),
        },
    )

    # No results returned
    send_message(
        read_until_client.connection,
        "ReadFish is disconnected from this device. Sequencing will proceed normally.",
        Severity.WARN,
    )


if __name__ == "__main__":
    main()



"""
Iteralign-centrifuge.
Design spec.
    1. Grab fastq files from a location.
    2. Classify reads that uniquely align to a reference genome.
    3. Generate a minimap2 index from the previously found reference genomes.
    4. Align the files ONCE to a reference.
    5. Calculate cumulative coverage information.
    6. Write a list of genomes that are covered at a particular threshold.
    7. Rinse and repeat
"""
import logging
import logging.handlers
import os
import subprocess
import sys
import threading
import time

from io import StringIO, BytesIO
import gzip

import argparse
from pathlib import Path
import shutil
import urllib.request as request
import urllib.error as url_error
from contextlib import closing

import pandas as pd
import toml
from watchdog.events import FileSystemEventHandler
from watchdog.observers.polling import PollingObserver as Observer

from ru.arguments import get_parser
from ru.utils import nice_join, print_args, send_message, Severity
from read_until_api_v2.load_minknow_rpc import get_rpc_connection, parse_message

from Bio import SeqIO
from collections import defaultdict

DEFAULT_SERVER_HOST = "127.0.0.1"
DEFAULT_LOG_FORMAT = "%(asctime)s - %(name)-20s - %(message)s"
LOG_LEVELS = ("debug", "info", "warning", "error", "critical")
DEFAULT_COVERAGE_DEPTH = 30
DEFAULT_PERCENTAGE_COVERED = 0.99
DEFAULT_CORES = 2

DEFAULT_THRESHOLD = 2000
DEFAULT_PREFIX = "iteraligncentrifuge"
DEFAULT_REJECT = "reject.tsv"
DEFAULT_CREPORT = "centrifuge_report.tsv"
DEFAULT_READS = "out.tsv"
DEFAULT_MINDEX = "mindex.mmi"
DEFAULT_GENOME = "genome.fna.gz"
DEFAULT_TIDFILE = "taxids.toml"
DEFAULT_COVERAGE_FILE = "coverage.tsv"
DEFAULT_SEQUENCE_LENGTH = 100000




def get_fq(s, pattern="*.fq"):
    return [str(f) for f in Path(s).rglob(pattern)]


_help = "ReadFish and Run Until, using centrifuge"
_cli = (
    (
        "--host",
        dict(
            metavar="HOST",
            help="MinKNOW server host",
            default=DEFAULT_SERVER_HOST,
        ),
    ),
    (
        "--watch",
        dict(
            metavar="FOLDER",
            help="Top Level Folder containing fastq reads.",
            default=None,
        ),
    ),
    (
        "--percent",
        dict(
            metavar="PERCENT",
            help="Default percent of target covered at given depth (default {})".format(DEFAULT_PERCENTAGE_COVERED),
            default=DEFAULT_PERCENTAGE_COVERED,
            type=float,
        ),
    ),
    (
        "--device",
        dict(
            required=True,
            action="store",
            help="The sequencing position being addressed"
        ),
    ),
    (
        "--depth",
        dict(
            metavar="DEPTH",
            help="Desired coverage depth (default {})".format(DEFAULT_COVERAGE_DEPTH),
            default=DEFAULT_COVERAGE_DEPTH,
            type=int,
        ),
    ),
    (
        "--threads",
        dict(
            metavar="THREADS",
            help="Set the number of default threads to use for threaded tasks (default {})".format(DEFAULT_CORES),
            default=DEFAULT_CORES,
            type=int,
        ),
    ),
    (
        "--log-level",
        dict(
            metavar="LOG-LEVEL",
            action="store",
            default="info",
            choices=LOG_LEVELS,
            help="One of: {}".format(nice_join(LOG_LEVELS)),
        ),
    ),
    (
        "--log-format",
        dict(
            metavar="LOG-FORMAT",
            action="store",
            default=DEFAULT_LOG_FORMAT,
            help="A standard Python logging format string (default: {!r})".format(
                DEFAULT_LOG_FORMAT.replace("%", "%%")
            ),
        ),
    ),
    (
        "--log-file",
        dict(
            metavar="LOG-FILE",
            action="store",
            default=None,
            help="A filename to write logs to, or None to write to the standard stream (default: None)",
        ),
    ),
    (
        "--cindex",
        dict(
            required=True,
            action="store",
            help="The centrifuge index required for read alignment. Must only have the prefix of the index names before the full stop and number of each index",
        ),
    ),
    (
        "--readfile",
        dict(
            action="store",
            help="The name of the file that will contain the metrics for the reads analysed",
            default=DEFAULT_READS,
        ),
    ),
    (
        "--csummary",
        dict(
            required=True,
            action="store",
            help="The file path to the custom centrifuge summary file with the ftp path for the reference genomes",
        ),
    ),
    (
        "--gfasta",
        dict(
            action="store",
            help="The file name for the genome references downloaded after centrifuge identification",
            default=DEFAULT_GENOME,
        ),
    ),
    (
        "--creport",
        dict(
            action="store",
            help="The file name of the centrifuge report file 'centrifuge_report.tsv' made each iteration. Will be created in this scripts source directory",
            default=DEFAULT_CREPORT,
        ),
    ),
    (
        "--keepfiles",
        dict(
            action="store_true",
            help="Set argument to retain the files generated by the centrifuge section of the script"
        ),
    ),
    (
        "--reject",
        dict(
            action="store",
            help="The file name containing reads not utilised within the 'call' file and the reason why.",
            default=DEFAULT_REJECT
        ),
    ),
    (
        "--mindex",
        dict(
            action="store",
            help="The file name where the minimap index will be stored",
            default=DEFAULT_MINDEX,
        ),
    ),
    (
        "--tidfile",
        dict(
            action="store",
            help="The file name where the taxIDs found with centrifuge are stored with the iteration they were found",
            default=DEFAULT_TIDFILE,
        ),
    ),
    (
        "--prefix",
        dict(
            action="store",
            help="The file path and prefix for all files generated by this script will generate",
            default=DEFAULT_PREFIX,
        ),
    ),
    (
        "--path",
        dict(
            required=True,
            action="store",
            help="The directory path where all files generated by this script will be stored"
        ),
    ),
    (
        "--threshold",
        dict(
            action="store",
            help="The threshold number of reads that are classified before a reference genome is added to the mmi",
            default=DEFAULT_THRESHOLD,
            type=int,
        ),
    ),
    (
        "--plasmids",
        dict(
            action="store",
            help="The path and file name of the plasmid file that will be used by the script",
        ),
    ),
    (
        "--references",
        dict(
            action="store",
            help="At least 1 taxID must be provided to be downloaded for mmi generation.",
            nargs="+",
            type=int,
        ),
    ),
    (
        "--seqlength",
        dict(
            action="store",
            help="The threshold length of an assembly to be incorporated into the reference file",
            default=DEFAULT_SEQUENCE_LENGTH,
            type=int,
        ),
    ),
    (
        "--coveragefile",
        dict(
            action="store",
            help="the file suffix containing 3 tab separated columns: iteration, referenceid, coverage (percent)",
            default=DEFAULT_COVERAGE_FILE,
        ),
    ),
    (
        "--toml",
        dict(
            metavar="TOML",
            required=True,
            help="The magic TOML file that will save your life?",
            # type=toml.load,
        ),
    ),
)


def file_dict_of_folder_simple(path, args, logging, fastqdict):
    logger = logging.getLogger("ExistingFileProc")

    file_list_dict = dict()

    counter = 0

    if os.path.isdir(path):

        logger.info("caching existing fastq files in: %s" % (path))

        for path, dirs, files in os.walk(path):

            for f in files:

                counter += 1

                if f.endswith(".fastq") or f.endswith(".fastq.gz") or f.endswith(".fq") or f.endswith(".fq.gz"):

                    logger.debug("Processing File {}\r".format(f))
                    filepath = os.path.join(path, f)
                    file_list_dict[filepath] = os.stat(filepath).st_mtime

    logger.info("processed %s files" % (counter))

    logger.info("found %d existing fastq files to process first." % (len(file_list_dict)))

    return file_list_dict


def url_list_generation(args,difference_set):
    # read in the file with taxIDs and their corresponding reference genome
    summary_centrifuge = pd.read_csv(args.csummary,
                                     sep="\t",
                                     )

    # create a set of ftp urls based on matching taxIDs between 'difference_set' and 'summary_centrifuge''
    url_list = list(summary_centrifuge[summary_centrifuge["taxid"].isin(difference_set)]["fasta_path"])

    return url_list


def download_references(args, url_list, taxid_set):

    lengths = {}
    for link in url_list:
        print("Attempting to download: {}".format(link), end="\n")
        try:
            response = request.urlopen(link)
        except url_error.URLError as e:
            print(e)
            print("Closing script")
            break
        compressed_file = BytesIO(response.read())
        with gzip.open(compressed_file, "rt") as fh, gzip.open(args.path + args.prefix + args.gfasta, "at") as fasta_seq:
            for seq_record in SeqIO.parse(fh, "fasta"):
                if len(seq_record) > args.seqlength:
                    lengths[seq_record.id] = len(seq_record)
                    SeqIO.write(seq_record, fasta_seq, "fasta")

    if args.plasmids:
        r = ("name", "path")
        logging.info("Obtaining the plasmids for the following taxids: {}".format(taxid_set))
        with open(args.csummary) as f:
            d = {int(x[0]): dict(zip(r, x[1:])) for i, l in enumerate(f) for x in (l.strip().split("\t"),)
                 if i > 0}

            for taxid in taxid_set:
                with gzip.open(args.plasmids, "rt") as fh, gzip.open(args.path + args.prefix + args.gfasta, "at") as fasta_seq:
                    for seq_record in SeqIO.parse(fh, "fasta"):
                        if d[taxid]["name"] in seq_record.description and len(seq_record) > args.seqlength:
                            lengths[seq_record.id] = len(seq_record)
                            SeqIO.write(seq_record, fasta_seq, "fasta")

    logging.info("Genome file generated")

    return lengths


def generate_mmi(args, counter):
    args.toml['conditions']['reference'] = os.path.abspath("{}{}_{}_{}".format(args.path,args.prefix,counter,args.mindex))
    minimap_cmd = ["minimap2", "-x", "map-ont", "-d", args.toml['conditions']['reference'], args.path + args.prefix + args.gfasta]

    # show the minimap command in the terminal
    logging.info(" ".join(minimap_cmd))

    # subprocess the 'minimap_cmd'
    minimap_db = subprocess.Popen(minimap_cmd,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE,
                                  stdin=subprocess.PIPE,
                                  universal_newlines=True,
                                  )
    out, err = minimap_db.communicate()

    minimap_db.stdout.close()


def parse_fastq_file(fastqfileList, args, logging, length_dict, taxID_set, counter, coverage_sum, connection):
    logger = logging.getLogger("ParseFastq")
    logger.info(fastqfileList)
    logger.info(args.toml['conditions']['reference'])
    with open(os.devnull, 'w') as devnull:

        # convert the 'fastqfileList' into a string valid for the list of fastq files to be read by centrifuge
        fastq_str = ",".join(fastqfileList)

        # centrifuge command to classify reads in the fastq files found by watchdog
        centrifuge_cmd = "centrifuge -p {} -x {} -q {}".format(args.threads,
                                                               args.cindex,
                                                               fastq_str
                                                               )

        # show what the centrifuge command in the terminal
        logging.info(centrifuge_cmd)

        # start time of centrifuge to track the time centrifuge requires to classify reads
        centrifuge_start_time = time.time()

        # subprocess for 'centrifuge_cmd'
        proc = subprocess.Popen(
            centrifuge_cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            stdin=subprocess.PIPE,
            shell=True,
            # Aliased by `text=True` in 3.7
            universal_newlines=True,
        )
        out, err = proc.communicate()
        proc.stdout.close()

        # confirm that the centrifuge pipeline has finished and show the time of execusion
        logging.info("Post centrifuge run: {} seconds".format(time.time()-centrifuge_start_time))

        output_fields = ["readID", "seqID", "taxID", "hitLength", "queryLength", "numMatches"]

        # create the DataFrame from the centrifuge output using 'output_fields' as the column headers
        out_df = pd.read_csv(StringIO(out),
                             sep="\t",
                             usecols=output_fields
                             )

        # create a dataframe from the args.creport 'centrifuge_report.tsv'
        report_df = pd.read_csv(args.creport,
                                sep="\t",
                                usecols=["name", "genomeSize", "taxID"],
                                )

        # merge both dataframes together
        new_df = pd.merge(out_df, report_df, on="taxID")

        # only reads that uniquely align to a species will be used, with reads that numMatches != 1 being filtered
        reject_df = new_df[new_df.numMatches != 1]
        reject_df["reason"] = "Multiple reference genome alignments"
        # reject_df.loc[:, "reason"] = "Multiple reference genome alignments"

        # genomeSize == 0 infer a read classification above the species taxon and are therefore removed
        intermediate_df = new_df[new_df.genomeSize == 0]
        intermediate_df = intermediate_df[intermediate_df.numMatches == 1]
        intermediate_df["reason"] = "Read aligns to non-species taxon"

        # log the reads that were removed into a seperate file
        reject_df = reject_df.append(intermediate_df)
        if os.path.isfile(args.path + args.prefix + args.reject):
            reject_df.to_csv(args.path + args.prefix + args.reject,
                             sep="\t",
                             mode="a",
                             index=False,
                             header=None,
                             )
        else:
            reject_df.to_csv(args.path + args.prefix + args.reject,
                             sep="\t",
                             mode="a",
                             index=False,
                             header=True,
                             )

        # reads that uniquely align to one species are logged into a file
        new_df = new_df[new_df.numMatches == 1]
        if os.path.isfile(args.path + args.prefix + args.readfile):
            new_df.to_csv(args.path + args.prefix + args.readfile,
                          sep="\t",
                          mode="a",
                          index=None,
                          header=None,
                          )
        else:
            new_df.to_csv(args.path + args.prefix + args.readfile,
                          sep="\t",
                          mode="a",
                          index=None,
                          header=True,
                          )
        logging.info("reject file made")

        # read in all valid reads back into memory from all current iterations
        all_reads_df = pd.read_csv(args.path + args.prefix + args.readfile,
                                   sep="\t"
                                   )

        # count the number of reads that were clasified to all taxIDs found within 'all_reads_df' and only keep the taxIDs with a count above args.threshold
        taxid_count = all_reads_df.groupby("taxID").agg({"taxID": "count"})
        taxid_result_set = set(taxid_count[taxid_count["taxID"].ge(args.threshold)].index.values.tolist())
        # call in the taxIDs that already have a reference genome in the mmi
        downloaded_set = taxID_set

        # if taxIDs were found above args.threshold
        if taxid_result_set:
            logging.info("taxIDs found above specified threshold")
            # locate new taxIDs that were not found in previous iterations
            difference_set = taxid_result_set - downloaded_set
            downloaded_set |= taxid_result_set

            # if novel taxIDs were found
            if difference_set:
                with open(args.path + args.prefix + args.tidfile, "w+") as f:
                    d = toml.load(args.path + args.prefix + args.tidfile)
                    if os.path.isfile(args.tidfile):
                        d["taxid"] = {"iteration.{}".format(counter): difference_set}
                    else:
                        d = {"taxid": {"iteration.{}".format(counter): difference_set}}
                    toml.dump(d, f)
                f.close()

                logging.info("new taxids found: {}".format(difference_set))
                logging.info("Downloading reference genomes")

                url_list = url_list_generation(args, difference_set)

                # download the reference genomes into a single file

                length_dict.update(download_references(args, url_list, difference_set))

                logging.info("Generating mmi")

                generate_mmi(args, counter)

                update_message = "Updated the minimap MMI to {}".format(args.toml['conditions']['reference'])
                logging.info(update_message)
                if not args.simulation:
                    #send_message_port(update_message, args.host, messageport)
                    send_message(connection, update_message, Severity.WARN)


            else:
                # show in terminal that no new taxIDs were found
                logging.info("No new taxIDs were identified")
        else:
            logging.info("No taxIDs found above threshold.")
        # all of the new code has been parsed, the rest follows standard 'iteralign.py' script
        logging.info("new code parsing completed\n\ncommencing iteralign")

        minimapcmd = ["minimap2","-ax","map-ont","-t {}".format(args.threads),args.toml['conditions']['reference']] #" ".join(fastqfilelist)]
        minimapcmd.extend(fastqfileList)
        logging.info(" ".join(minimapcmd))
        minimapoutput = subprocess.Popen(minimapcmd, stdout=subprocess.PIPE,stderr=devnull)

        samcmd = ["samtools","view", "-bS"]
        samoutput = subprocess.Popen(samcmd, stdin=minimapoutput.stdout, stdout=subprocess.PIPE, stderr=devnull)
        #samsortcmd = ["samtools", "sort", "-@2", "-o", "sortedbam.bam"]
        samsortcmd = ["samtools", "sort", "-@{}".format(args.threads)]
        samsortoutput = subprocess.Popen(samsortcmd, stdin=samoutput.stdout, stdout=subprocess.PIPE, stderr=devnull)
        samdepthcmd = ["samtools", "depth", "/dev/stdin"]
        samdepthoutput = subprocess.Popen(samdepthcmd, stdin=samsortoutput.stdout,stdout=subprocess.PIPE, stderr=devnull, universal_newlines=True)
        minimapoutput.stdout.close()
        samoutput.stdout.close()
        samsortoutput.stdout.close()

        iter_depth = (l.strip().split("\t") for l in samdepthoutput.stdout)
        parse_iter = ((x[0], int(x[-1])) for x in iter_depth)

        d = defaultdict(int)
        d.update(coverage_sum)
        for name, depth in parse_iter:
            d[name] += depth

        depth_dict = {k: d[k]/length_dict[k] for k in length_dict.keys() & d}

        with open(args.path + args.prefix + args.coveragefile, "a") as fh:
            for k, v in depth_dict.items():
                fh.write("{}\t{}\t{}\n".format(counter, k, v))

        targets = [k for k, v in depth_dict.items() if v > args.depth]

        logging.info(targets)

        counter += 1

        logging.info("Finished processing {}.".format(" ".join(fastqfileList)))

    return targets, downloaded_set, counter, d


def write_new_toml(args,targets):
    for k in args.toml["conditions"].keys():
        curcond = args.toml["conditions"].get(k)
        if isinstance(curcond,dict):

            #newtargets = targets
            #newtargets.extend(curcond["targets"])

            #newtargets = list(dict.fromkeys(newtargets))
            #curcond["targets"]=list(set(newtargets))
            curcond["targets"]=targets

    with open("{}_live".format(args.tomlfile), "w") as f:
        toml.dump(args.toml,f)


class FastqHandler(FileSystemEventHandler):

    def __init__(self, args, logging, messageport, rpc_connection):
        self.args = args
        self.messageport = messageport
        self.connection = rpc_connection
        self.logger = logging.getLogger("FastqHandler")
        self.running = True
        self.fastqdict = dict()
        self.creates = file_dict_of_folder_simple(self.args.watch, self.args, logging,
                                                  self.fastqdict)
        self.t = threading.Thread(target=self.processfiles)

        try:
            self.t.start()
        except KeyboardInterrupt:
            self.t.stop()
            raise

    def processfiles(self):
        self.logger.info("Process Files Inititated")
        self.counter = 1
        self.targets = []
        self.masterdf = pd.DataFrame(columns=['seqid', 'position', 'coverage'])
        self.taxid_entries = 0
        self.downloaded_set = set()
        self.length_dict = {}
        self.coverage_sum = {}

        if self.args.references:
            logging.info("References argument provided. Will download references genomes.")
            self.downloaded_set = set(self.args.references)
            logging.info(self.downloaded_set)
            self.url_list = url_list_generation(self.args, self.args.references)
            self.length_dict.update(download_references(self.args, self.url_list, self.downloaded_set))
            generate_mmi(self.args, self.counter)


        while self.running:
            currenttime = time.time()
            # for fastqfile, createtime in tqdm(sorted(self.creates.items(), key=lambda x: x[1])):
            fastqfilelist = list()
            for fastqfile, createtime in sorted(self.creates.items(), key=lambda x: x[1]):

                delaytime = 0

                # file created 5 sec ago, so should be complete. For simulations we make the time longer.
                if (int(createtime) + delaytime < time.time()):
                    self.logger.info(fastqfile)
                    del self.creates[fastqfile]
                    self.counter += 1
                    fastqfilelist.append(fastqfile)

                    # print (fastqfile,md5Checksum(fastqfile), "\n\n\n\n")
            # as long as there are files within the args.watch directory to parse
            if fastqfilelist:
                print(self.downloaded_set)
                targets, self.downloaded_set, self.taxid_entries, self.coverage_sum = parse_fastq_file(fastqfilelist, self.args, logging, self.length_dict, self.downloaded_set, self.taxid_entries, self.coverage_sum, self.connection)
                print(targets)
                print(self.targets)

                if len(targets) > len(self.targets):
                    updated_targets = set(targets) - set(self.targets)
                    update_message = "Updating targets with {}".format(nice_join(updated_targets, conjunction="and"))
                    self.logger.info(update_message)
                    if not self.args.simulation:
                        #send_message_port(update_message, self.args.host, self.messageport)
                        send_message(self.connection, update_message, Severity.WARN)
                    write_new_toml(self.args, targets)
                    self.targets = []
                    self.targets = targets.copy()

                if self.masterdf.shape[0] > 0 and self.masterdf.shape[0] == len(self.targets):
                    # Every target is covered at the desired coverage level.
                    self.logger.info("Every target is covered at at least {}x".format(self.args.depth))
                    if not self.args.simulation:
                        self.connection.protocol.stop_protocol()
                        #send_message_port(
                        #    "Iter Align has stopped the run as all targets should be covered by at least {}x".format(
                        #        self.args.depth), self.args.host, self.messageport)
                        send_message(self.connection, "Iter Align has stopped the run as all targets should be covered by at least {}x".format(
                                self.args.depth), Severity.WARN)

                # parse_fastq_file(fastqfile, self.rundict, self.fastqdict, self.args, self.header, self.MinotourConnection)

                # self.args.files_processed += 1

                if currenttime + 5 > time.time():
                    time.sleep(5)

    def on_created(self, event):
        """Watchdog counts a new file in a folder it is watching as a new file"""
        """This will add a file which is added to the watchfolder to the creates and the info file."""
        # if (event.src_path.endswith(".fastq") or event.src_path.endswith(".fastq.gz")):
        #     self.creates[event.src_path] = time.time()

        # time.sleep(5)
        if (event.src_path.endswith(".fastq") or event.src_path.endswith(".fastq.gz") or event.src_path.endswith(
                ".fq") or event.src_path.endswith(".fq.gz")):
            self.logger.info("Processing file {}".format(event.src_path))
            self.creates[event.src_path] = time.time()

    def on_modified(self, event):
        if (event.src_path.endswith(".fastq") or event.src_path.endswith(".fastq.gz") or event.src_path.endswith(
                ".fq") or event.src_path.endswith(".fq.gz")):
            self.logger.info("Processing file {}".format(event.src_path))
            self.logger.debug("Modified file {}".format(event.src_path))
            self.creates[event.src_path] = time.time()

    def on_moved(self, event):
        if any((event.dest_path.endswith(".fastq"), event.dest_path.endswith(".fastq,gz"),
                event.dest_path.endswith(".gq"), event.dest_path.endswith(".fq.gz"))):
            self.logger.info("Processing file {}".format(event.dest_path))
            self.logger.debug("Modified file {}".format(event.dest_path))
            self.creates[event.dest_path] = time.time()


def main():
    sys.exit(
        "This entry point is deprecated, please use 'readfish centrifuge' instead"
    )

def run(parser, args):
    # new code block: change the reference path within the args.toml file into the args.mindex path
    d = toml.load(args.toml)

    print(d["conditions"]["reference"])
    args.tomlfile = args.toml
    args.toml = toml.load(args.toml)
    print(args)

    # TODO: Move logging config to separate configuration file
    # set up logging to file
    logging.basicConfig(level=logging.DEBUG,
                        format='%(levelname)s::%(asctime)s::%(name)s::%(message)s',
                        filename=args.log_file,
                        filemode='w')

    # define a Handler that writes INFO messages or higher to the sys.stderr
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)

    # set a format which is simpler for console use
    formatter = logging.Formatter('%(name)-15s: %(levelname)-8s %(message)s')
    console.setFormatter(formatter)

    # add the handler to the root logger
    logging.getLogger('').addHandler(console)

    # Start by logging sys.argv and the parameters used
    logger = logging.getLogger("Manager")
    logger.info(" ".join(sys.argv))
    print_args(args, logger=logger)

    logger.info("Initialising iterAlign.")

    logger.info("Setting up FastQ monitoring.")

    #### Check if a run is active - if not, wait.

    args.simulation = True
    connection = None

    #set default message severity level.
    severity = 2

    if args.watch is None:
        args.simulation = False
        logger.info("Creating rpc connection for device {}.".format(args.device))
        try:
            connection, messageport = get_rpc_connection(args.device)
        except ValueError as e:
            print(e)
            sys.exit(1)

        #send_message_port("Iteralign Connected to MinKNOW", args.host, messageport)
        send_message(connection, "Iteralign Connected to MinKNOW.", Severity.WARN)

        logger.info("Loaded RPC")
        while parse_message(connection.acquisition.current_status())['status'] != "PROCESSING":
            time.sleep(1)
        ### Check if we know where data is being written to , if not... wait
        args.watch = parse_message(connection.acquisition.get_acquisition_info())['config_summary'][
            'reads_directory']

    else:
        messageport = ""

    event_handler = FastqHandler(args, logging, messageport, connection)
    # This block handles the fastq
    observer = Observer()
    observer.schedule(event_handler, path=args.watch, recursive=True)
    observer.daemon = True

    try:

        observer.start()
        logger.info("FastQ Monitoring Running.")
        while 1:
            time.sleep(1)

    except KeyboardInterrupt:

        logger.info("Exiting - Will take a few seconds to clean up!")

        observer.stop()
        observer.join()

        if args.keepfiles:
            logging.info("The 'keepfiles' argument was set, files generated by classifier have been retained")
        else:
            if os.path.isdir(args.path):
                for path, dirs, files in os.walk(args.path):
                    for f in files:
                        if f.startswith(args.prefix):
                            os.unlink(f)
                            logging.info("file removed: {}".format(f))

            if os.path.isdir("./"):
                for path, dirs, files in os.walk("./"):
                    for f in files:
                        if f.endswith(args.creport):
                            os.unlink(f)
                            logging.info("file removed: {}".format(f))

        logging.info("All files generated by classifier have been removed.")

        os._exit(0)


if __name__ == "__main__":
    main()



"""unblock_all.py

ReadUntil implementation that will only unblock reads. This should result in
a read length histogram that has very short peaks (~280-580bp) as these are the
smallest chunks that we can acquire. If you are not seeing these peaks, the
`split_reads_after_seconds` parameter in the configuration file may need to be
edited to 0.2-0.4:
(<MinKNOW_folder>/ont-python/lib/python2.7/site-packages/bream4/configuration)
"""
# Core imports
import functools
import logging
import sys
import time
from timeit import default_timer as timer

# Read Until imports
import read_until_api_v2 as read_until
from read_until_api_v2.utils import run_workflow
from ru.arguments import BASE_ARGS
from ru.utils import print_args
from ru.utils import send_message, Severity


_help = "Unblock all reads"
_cli = BASE_ARGS


def simple_analysis(client, batch_size=512, throttle=0.1, unblock_duration=0.1):
    """Analysis function

    Parameters
    ----------
    client : read_until.ReadUntilClient
        An instance of the ReadUntilClient object
    batch_size : int
        The number of reads to be retrieved from the ReadUntilClient at a time
    throttle : int or float
        The number of seconds interval between requests to the ReadUntilClient
    unblock_duration : int or float
        Time, in seconds, to apply unblock voltage

    Returns
    -------
    None
    """
    logger = logging.getLogger(__name__)
    send_message(client.connection, "ReadFish sending Unblock All Messages. All reads will be prematurely truncated. This will affect a live sequencing run.",
                 Severity.WARN)
    while client.is_running:

        r = 0
        t0 = timer()

        for r, (channel, read) in enumerate(
                client.get_read_chunks(
                    batch_size=batch_size,
                    last=True,
                ),
                start=1,
        ):
            # pass
            client.unblock_read(channel, read.number, read_id=read.id, duration=unblock_duration)
            client.stop_receiving_read(channel, read.number)

        t1 = timer()
        if r:
            logger.info("Took {:.6f} for {} reads".format(t1-t0, r))
        # limit the rate at which we make requests
        if t0 + throttle > t1:
            time.sleep(throttle + t0 - t1)
    else:
        send_message(client.connection, "ReadFish Unblock All Disconnected.", Severity.WARN)
        logger.info("Finished analysis of reads as client stopped.")


def main():
    sys.exit(
        "This entry point is deprecated, please use 'readfish unblock-all' instead"
    )


def run(parser, args):
    # TODO: Move logging config to separate configuration file
    # TODO: use setup_logger here instead?
    # set up logging to file for DEBUG messages and above
    logging.basicConfig(
        level=logging.DEBUG,
        # TODO: args.log_format
        format="%(levelname)s %(asctime)s %(name)s %(message)s",
        filename=args.log_file,
        filemode="w",
    )

    # define a Handler that writes INFO messages or higher to the sys.stderr
    console = logging.StreamHandler()
    console.setLevel(logging.INFO)

    # set a format which is simpler for console use
    formatter = logging.Formatter(args.log_format)
    console.setFormatter(formatter)

    # add the handler to the root logger
    logging.getLogger("").addHandler(console)

    # Start by logging sys.argv and the parameters used
    logger = logging.getLogger("Manager")
    # logger = setup_logger(__name__, args.log_format, log_file=args.log_file, level=logging.INFO)
    logger.info(" ".join(sys.argv))
    print_args(args, logger=logger)

    read_until_client = read_until.ReadUntilClient(
        mk_host=args.host,
        mk_port=args.port,
        device=args.device,
        # one_chunk=args.one_chunk,
        filter_strands=True,
        # TODO: test cache_type by passing a function here
        cache_type=args.read_cache,
        cache_size=args.cache_size,
    )

    analysis_worker = functools.partial(
        simple_analysis,
        client=read_until_client,
        batch_size=args.batch_size,
        throttle=args.throttle,
        unblock_duration=args.unblock_duration,
    )

    results = run_workflow(
        client=read_until_client,
        partial_analysis_func=analysis_worker,
        n_workers=args.workers,
        run_time=args.run_time,
        runner_kwargs={
            # "min_chunk_size": args.min_chunk_size,
            "first_channel": args.channels[0],
            "last_channel": args.channels[-1],
        },
    )
    # No results returned


if __name__ == "__main__":
    main()



import sys
import argparse
import textwrap

from ru.utils import get_run_info, describe_experiment, Severity
from ru.basecall import Mapper


_help = "ReadFish TOML Validator"
_cli = (
    ("toml", dict(help="TOML file to validate")),
)


class colour:
   GREEN = '\033[92m'
   YELLOW = '\033[93m'
   RED = '\033[91m'
   BOLD = '\033[1m'
   END = '\033[0m'


def printer(s, sev, **kwargs):
    if sev == Severity.INFO:
        print(s, **kwargs)
    elif sev == Severity.WARN:
        print("{}{}{}".format(colour.YELLOW, s, colour.END), **kwargs)
    elif sev == Severity.ERROR:
        print("{}{}{}".format(colour.RED, s, colour.END), **kwargs)


def except_hook(type, value, traceback):
    print(value, file=sys.stderr)


def run(parser, args):
    # Catch exceptions and only print error line
    sys.excepthook = except_hook

    # Run load config to validate
    run_info, conditions, reference, caller_settings = get_run_info(args.toml)
    print("😻 Looking good!", file=sys.stdout)
    print("Generating experiment description - please be patient!", file=sys.stdout)
    mapper = Mapper(reference)
    for message, sev in describe_experiment(conditions, mapper):
        printer(textwrap.fill(message), sev, file=sys.stdout, end="\n\n")


if __name__ == "__main__":
    main()



import argparse
import sys

import read_until_api_v2.read_cache as RC
from ru.utils import nice_join

# TODO: Add prefix parameter that is applied to all log files

DEFAULT_SERVER_HOST = "127.0.0.1"
DEFAULT_SERVER_PORT = 9501
DEFAULT_WORKERS = 1
DEFAULT_LOG_FORMAT = "%(asctime)s %(name)s %(message)s"
DEFAULT_LOG_LEVEL = "info"
DEFAULT_CHANNELS = [1, 512]
DEFAULT_DELAY = 0
DEFAULT_RUN_TIME = 172800
DEFAULT_UNBLOCK = 0.1
DEFAULT_CACHE_SIZE = 512
DEFAULT_BATCH_SIZE = 512
DEFAULT_THROTTLE = 0.1
DEFAULT_MIN_CHUNK = 2000
DEFAULT_LOG_PREFIX = ""

LOG_LEVELS = ("debug", "info", "warning", "error", "critical")
READ_CACHE = RC.__all__

if "ONTReadCache" in READ_CACHE:
    DEFAULT_READ_CACHE = "ONTReadCache"
else:
    try:
        DEFAULT_READ_CACHE = READ_CACHE[0]
    except IndexError:
        raise IndexError("No ReadCache classes found")

BASE_ARGS = (
    (
        "--host",
        dict(
            metavar="HOST",
            help="MinKNOW server host (default: {})".format(DEFAULT_SERVER_HOST),
            default=DEFAULT_SERVER_HOST,
        ),
    ),
    (
        "--port",
        dict(
            metavar="PORT",
            help="MinKNOW server port (default: {})".format(DEFAULT_SERVER_PORT),
            default=DEFAULT_SERVER_PORT,
        ),
    ),
    (
        "--device",
        dict(
            metavar="DEVICE",
            type=str,
            help="Name of the sequencing position e.g. MS29042 or X1 etc.",
            required=True,
        ),
    ),
    (
        "--experiment-name",
        dict(
            metavar="EXPERIMENT-NAME",
            type=str,
            help="Describe the experiment being run, enclose in quotes",
            required=True,
        ),
    ),
    (
        "--read-cache",
        dict(
            metavar="READ_CACHE",
            action="store",
            default=DEFAULT_READ_CACHE,
            choices=READ_CACHE,
            help="One of: {} (default: {})".format(
                nice_join(READ_CACHE), DEFAULT_READ_CACHE
            ),
        ),
    ),
    (
        "--workers",
        dict(
            metavar="WORKERS",
            type=int,
            help="Number of worker threads (default: {})".format(DEFAULT_WORKERS),
            default=DEFAULT_WORKERS,
        ),
    ),
    (
        "--channels",
        dict(
            metavar="CHANNELS",
            type=int,
            nargs=2,
            help="Channel range to use as a sequence, expects two integers "
                 "separated by a space (default: {})".format(
                    DEFAULT_CHANNELS
            ),
            default=DEFAULT_CHANNELS,
        ),
    ),
    (
        "--run-time",
        dict(
            metavar="RUN-TIME",
            type=int,
            help="Period (seconds) to run the analysis (default: {:,})".format(
                DEFAULT_RUN_TIME
            ),
            default=DEFAULT_RUN_TIME,
        ),
    ),
    (
        "--unblock-duration",
        dict(
            metavar="UNBLOCK-DURATION",
            type=int,
            help="Time, in seconds, to apply unblock voltage (default: {})".format(
                DEFAULT_UNBLOCK
            ),
            default=DEFAULT_UNBLOCK,
        ),
    ),
    (
        "--cache-size",
        dict(
            metavar="CACHE-SIZE",
            type=int,
            help="The size of the read cache in the ReadUntilClient (default: {:,})".format(
                DEFAULT_CACHE_SIZE
            ),
            default=DEFAULT_CACHE_SIZE,
        ),
    ),
    (
        "--batch-size",
        dict(
            metavar="BATCH-SIZE",
            type=int,
            help="The maximum number of reads to pull from the read cache (default: {:,})".format(
                DEFAULT_BATCH_SIZE
            ),
            default=DEFAULT_BATCH_SIZE,
        ),
    ),
    (
        "--throttle",
        dict(
            metavar="THROTTLE",
            type=float,
            help="Time interval, in seconds, between requests to the ReadUntilClient (default: {})".format(
                DEFAULT_THROTTLE
            ),
            default=DEFAULT_THROTTLE,
        ),
    ),
    (
        "--dry-run",
        dict(
            action="store_true",
            help="Run the ReadFish Until experiment without sending unblock commands",
        ),
    ),
    (
        "--log-level",
        dict(
            metavar="LOG-LEVEL",
            action="store",
            default=DEFAULT_LOG_LEVEL,
            choices=LOG_LEVELS,
            help="One of: {}".format(nice_join(LOG_LEVELS)),
        ),
    ),
    (
        "--log-format",
        dict(
            metavar="LOG-FORMAT",
            action="store",
            default=DEFAULT_LOG_FORMAT,
            help="A standard Python logging format string (default: {!r})".format(
                DEFAULT_LOG_FORMAT.replace("%", "%%")
            ),
        ),
    ),
    (
        "--log-file",
        dict(
            metavar="LOG-FILE",
            action="store",
            default=None,
            help="A filename to write logs to, or None to write to the standard stream (default: None)",
        ),
    ),
)


def get_parser(extra_args=None, file=None, default_args=None):
    """Generic argument parser for ReadFish scripts

    Parameters
    ----------
    extra_args : Tuple[Tuple[str, dict], ...]
        Extra arguments to append onto the base arguments
    file : str
        Optional. __file__ from the python script, used for program string
    default_args : Tuple[Tuple[str, dict], ...]
        Arguments that form the base requirements for all ReadFish scripts

    Returns
    -------
    parser : argparse.ArgumentParser
        The argparse parser, used for raising parser errors manually
    arguments : argparse.ArgumentParser().parse_args()
        The parsed arguments
    """
    if default_args is None:
        args = BASE_ARGS
    else:
        args = default_args

    if extra_args is not None:
        args = args + extra_args

    if file is None:
        prog_string = "ReadFish API: {}".format(sys.argv[0].split("/")[-1])
    else:
        prog_string = "ReadFish API: {} ({})".format(sys.argv[0].split("/")[-1], file)

    parser = argparse.ArgumentParser(prog_string)
    for arg in args:
        flags = arg[0]
        if not isinstance(flags, tuple):
            flags = (flags,)
        parser.add_argument(*flags, **arg[1])

    return parser, parser.parse_args()



from setuptools import setup, find_packages
from os import path

PKG_NAME = "readfish"
MOD_NAME = "ru"

DESCRIPTION = """# <img src="https://raw.githubusercontent.com/LooseLab/ru/rename_cli/examples/images/readfish_logo.jpg">

Installation
---

This toolkit currently requires MinKNOW (minknow-core v3.6) to be installed and `read_until_api_v2` to be installed
separately. We recommend installing in a virtual environment as so:

```bash
# Make a virtual env
python3 -m venv readfish
source ./readfish/bin/activate
pip install git+https://github.com/LooseLab/read_until_api_v2@master
pip install readfish
```

Usage
---
```bash
# check install
$ readfish
usage: readfish [-h] [--version]
                {targets,align,centrifuge,unblock-all,validate,summary} ...

positional arguments:
  {targets,align,centrifuge,unblock-all,validate,summary}
                        Sub-commands
    targets             Run targeted sequencing
    align               ReadFish and Run Until, using minimap2
    centrifuge          ReadFish and Run Until, using centrifuge
    unblock-all         Unblock all reads
    validate            ReadFish TOML Validator
    summary             Summary stats from FASTQ files

optional arguments:
  -h, --help            show this help message and exit
  --version             show program's version number and exit

See '<command> --help' to read about a specific sub-command.

# example run command - change arguments as necessary:
$ readfish targets --experiment-name "Test run" --device MN17073 --toml example.toml --log-file RU_log.log
```
"""

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md")) as fh, open(path.join(here, "requirements.txt")) as req:
    install_requires = [pkg.strip() for pkg in req]

__version__ = ""
exec(open("{}/_version.py".format(MOD_NAME)).read())

setup(
    name=PKG_NAME,
    version=__version__,
    author="Alexander Payne",
    author_email="alexander.payne@nottingham.ac.uk",
    description="Adaptive sampling toolkit for MinION",
    long_description=DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/LooseLab",
    packages=find_packages(exclude=["*.test", "*.test.*", "test.*", "test"]),
    entry_points={
        "console_scripts": [
            "ru_validate={}.validate:main".format(MOD_NAME),
            "ru_generators={}.ru_gen:main".format(MOD_NAME),
            "ru_summarise_fq={}.summarise_fq:main".format(MOD_NAME),
            "ru_iteralign={}.iteralign:main".format(MOD_NAME),
            "ru_iteralign_centrifuge={}.iteralign_centrifuge:main".format(MOD_NAME),
            "ru_unblock_all={}.unblock_all:main".format(MOD_NAME),
            "readfish={}.cli:main".format(MOD_NAME),
        ],
    },
    install_requires=install_requires,
    include_package_data=True,
    python_requires=">=3.5",
)
